from collections.abc import Iterable

import constructs
from aibs_informatics_aws_utils.constants.efs import (
    EFS_SCRATCH_PATH,
    EFS_SHARED_PATH,
    EFS_TMP_PATH,
)
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_batch as batch
from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_s3 as s3

from aibs_informatics_cdk_lib.common.aws.iam_utils import (
    BATCH_FULL_ACCESS_ACTIONS,
    batch_policy_statement,
)
from aibs_informatics_cdk_lib.constructs_.assets.code_asset_definitions import (
    AIBSInformaticsAssets,
)
from aibs_informatics_cdk_lib.constructs_.efs.file_system import (
    EFSEcosystem,
    MountPointConfiguration,
)
from aibs_informatics_cdk_lib.constructs_.service.compute import BatchCompute, LambdaCompute
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.informatics import (
    BatchInvokedLambdaFunction,
    DataSyncFragment,
    DemandExecutionFragment,
)
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.informatics.efs import (
    CleanFileSystemFragment,
    CleanFileSystemTriggerConfig,
    CleanFileSystemTriggerRuleConfig,
)
from aibs_informatics_cdk_lib.stacks.base import EnvBaseStack

DATA_SYNC_ASSET_NAME = "aibs_informatics_aws_lambda"

EFS_MOUNT_PATH = "/opt/efs"
EFS_VOLUME_NAME = "efs-file-system"


class DemandExecutionInfrastructureStack(EnvBaseStack):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str | None,
        env_base: EnvBase,
        vpc: ec2.Vpc,
        buckets: Iterable[s3.Bucket] | None = None,
        mount_point_configs: Iterable[MountPointConfiguration] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(scope, id, env_base, **kwargs)
        self.execution_compute = BatchCompute(
            self,
            id="demand",
            env_base=env_base,
            batch_name="demand",
            vpc=vpc,
            buckets=buckets,
            mount_point_configs=mount_point_configs,
        )
        self.infra_compute = LambdaCompute(
            self,
            id="demand-infra",
            env_base=env_base,
            vpc=vpc,
            batch_name="demand-infra",
            buckets=buckets,
            mount_point_configs=mount_point_configs,
            instance_role_policy_statements=[
                batch_policy_statement(env_base=env_base, actions=BATCH_FULL_ACCESS_ACTIONS),
            ],
        )


class DemandExecutionStack(EnvBaseStack):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        assets: AIBSInformaticsAssets,
        scaffolding_bucket: s3.Bucket,
        efs_ecosystem: EFSEcosystem,
        scaffolding_job_queue: batch.JobQueue | str,
        data_sync_job_queue: batch.JobQueue | str,
        execution_job_queue: batch.JobQueue | str,
        **kwargs,
    ):
        super().__init__(scope, id, env_base, **kwargs)

        self.efs_ecosystem = efs_ecosystem

        self.execution_job_queue = (
            execution_job_queue
            if isinstance(execution_job_queue, str)
            else execution_job_queue.job_queue_arn
        )
        self.data_sync_job_queue = (
            data_sync_job_queue
            if isinstance(data_sync_job_queue, str)
            else data_sync_job_queue.job_queue_arn
        )
        self.scaffolding_job_queue = (
            scaffolding_job_queue
            if isinstance(scaffolding_job_queue, str)
            else scaffolding_job_queue.job_queue_arn
        )

        self._assets = assets

        root_mount_point_config = MountPointConfiguration.from_access_point(
            self.efs_ecosystem.root_access_point, EFS_MOUNT_PATH
        )
        shared_mount_point_config = MountPointConfiguration.from_access_point(
            self.efs_ecosystem.shared_access_point, "/opt/shared", read_only=True
        )
        scratch_mount_point_config = MountPointConfiguration.from_access_point(
            self.efs_ecosystem.scratch_access_point, "/opt/scratch"
        )

        batch_invoked_lambda_fragment = BatchInvokedLambdaFunction.with_defaults(
            self,
            "batch-invoked-lambda",
            env_base=self.env_base,
            name="batch-invoked-lambda",
            job_queue=self.scaffolding_job_queue,
            bucket_name=scaffolding_bucket.bucket_name,
            mount_point_configs=[root_mount_point_config],
        )
        self.batch_invoked_lambda_state_machine = batch_invoked_lambda_fragment.to_state_machine(
            "batch-invoked-lambda-state-machine"
        )

        data_sync = DataSyncFragment(
            self,
            "data-sync",
            env_base=self.env_base,
            aibs_informatics_docker_asset=self._assets.docker_assets.AIBS_INFORMATICS_AWS_LAMBDA,
            batch_job_queue=self.data_sync_job_queue,
            scaffolding_bucket=scaffolding_bucket,
            mount_point_configs=[root_mount_point_config],
        )

        self.data_sync_state_machine = data_sync.to_state_machine("data-sync-v2")

        demand_execution = DemandExecutionFragment(
            self,
            "demand-execution",
            env_base=self.env_base,
            aibs_informatics_docker_asset=self._assets.docker_assets.AIBS_INFORMATICS_AWS_LAMBDA,
            scaffolding_bucket=scaffolding_bucket,
            scaffolding_job_queue=self.scaffolding_job_queue,
            batch_invoked_lambda_state_machine=self.batch_invoked_lambda_state_machine,
            data_sync_state_machine=self.data_sync_state_machine,
            shared_mount_point_config=shared_mount_point_config,
            scratch_mount_point_config=scratch_mount_point_config,
            tags={
                "ai:cost-allocation:aibs-informatics-service": "n/a",
                "ai:cost-allocation:aibs-informatics-workflow-type": "$.execution_type",
                "ai:cost-allocation:aibs-informatics-workflow-id": "$.execution_id",
            },
        )
        self.demand_execution_state_machine = demand_execution.to_state_machine("demand-execution")

        ## EFS Cleanup

        clean_fs = CleanFileSystemFragment(
            self,
            "clean-file-system",
            env_base=self.env_base,
            aibs_informatics_docker_asset=self._assets.docker_assets.AIBS_INFORMATICS_AWS_LAMBDA,
            batch_job_queue=self.execution_job_queue,
            scaffolding_bucket=scaffolding_bucket,
            mount_point_configs=[root_mount_point_config],
        )
        self.clean_file_system_state_machine = clean_fs.to_state_machine("clean-file-system")

        CleanFileSystemTriggerRuleConfig(
            rule_name="clean-file-system-trigger",
            trigger_configs=[
                CleanFileSystemTriggerConfig(
                    file_system=self.efs_ecosystem.file_system,
                    path=path,
                    days_since_last_accessed=days_since_last_accessed,
                    max_depth=max_depth,
                    min_depth=min_depth,
                    min_size_bytes_allowed=0,
                )
                for path, days_since_last_accessed, min_depth, max_depth in [
                    (EFS_TMP_PATH, 3.0, 1, 1),
                    (EFS_SCRATCH_PATH, 3.0, 1, 1),
                    (f"{EFS_SCRATCH_PATH}/tmp", 3.0, 1, 1),
                    (EFS_SHARED_PATH, 3.0, 1, 1),
                ]
            ],
        ).create_rule(self, clean_file_system_state_machine=self.clean_file_system_state_machine)



from enum import Enum
from typing import TypeVar

import constructs
from aibs_informatics_core.utils.os_operations import get_env_var

from aibs_informatics_cdk_lib.project.utils import _get_from_context

CDK_STACK_TARGET_ENV_VAR = "CDK_STACK_TARGET"

T = TypeVar("T", bound="CDKStackTargetBaseEnum")


class CDKStackTargetBaseEnum(Enum):
    """Base class for CDK stack target types

    Usage:
        class MyCDKStackTarget(str, CDKStackTargetBaseEnum):
            INFRA = "pipeline"

    """

    @classmethod
    def from_env(cls: type[T], default: str | T) -> T:
        target = get_env_var(CDK_STACK_TARGET_ENV_VAR)
        target = target or default
        return cls(target)

    @classmethod
    def from_context(
        cls: type[T],
        node: constructs.Node,
        default: str | T,
        context_keys: list[str] | None = None,
    ) -> T:
        """Resolves the CDK stack target type from context

        Args:
            cls (Type[T]): subclassed CDKStackTargetBase
            node (constructs.Node): cdk construct node
            default (str): default to use.
            context_keys (Optional[list[str]], optional): overrides for context names.
                Defaults to None.

        Returns:
            CDKStackTargetBase instance
        """
        context_keys = context_keys or ["target", "stack_target"]

        target = _get_from_context(node, context_keys) or default

        return cls(target)

    @classmethod
    def from_context_or_env(
        cls: type[T],
        node: constructs.Node,
        default: str | T,
        context_keys: list[str] | None = None,
    ) -> T:
        """Resolves the CDK stack target type from context or environment

        Order of resolution:
            1. CDK context value (specifying -c K=V)
            2. env variable
            3. default value ("dev")

        Args:
            cls (Type[T]): subclassed CDKStackTargetBase
            node (constructs.Node): cdk construct node
            default (str): default to use.
            context_keys (Optional[list[str]], optional): overrides for context names.
                Defaults to None.

        """

        return cls.from_context(
            node=node, default=cls.from_env(default), context_keys=context_keys
        )



__all__ = [
    "BatchInvokedExecutorFragment",
    "BatchInvokedLambdaFunction",
    "DataSyncFragment",
    "DistributedDataSyncFragment",
    "DemandExecutionFragment",
    "CleanFileSystemFragment",
    "CleanFileSystemTriggerConfig",
]

from aibs_informatics_cdk_lib.constructs_.sfn.fragments.informatics.batch import (
    BatchInvokedExecutorFragment,
    BatchInvokedLambdaFunction,
)
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.informatics.data_sync import (
    DataSyncFragment,
    DistributedDataSyncFragment,
)
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.informatics.demand_execution import (
    DemandExecutionFragment,
)
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.informatics.efs import (
    CleanFileSystemFragment,
    CleanFileSystemTriggerConfig,
)



"""Project configuration models for CDK applications.

This module provides Pydantic models for defining project, stage, and pipeline
configurations for CDK applications.
"""

__all__ = [
    "Env",
    "EnvBase",
    "EnvType",
    "CodePipelineSourceConfig",
    "CodePipelineNotificationsConfig",
    "GlobalConfig",
    "StageConfig",
    "PipelineConfig",
    "BaseProjectConfig",
    "ProjectConfig",
    "ConfigProvider",
]

import logging
from collections.abc import MutableMapping
from pathlib import Path
from typing import Annotated, Generic, Self, TypeVar

import yaml
from aibs_informatics_core.collections import DeepChainMap
from aibs_informatics_core.env import ENV_BASE_KEY, ENV_LABEL_KEY, ENV_TYPE_KEY, EnvBase, EnvType
from aibs_informatics_core.models.unique_ids import UniqueID
from aibs_informatics_core.utils.file_operations import find_paths
from aibs_informatics_core.utils.os_operations import expandvars
from aibs_informatics_core.utils.tools.dicttools import remove_null_values
from pydantic import BaseModel, PlainSerializer, PlainValidator, model_validator

UniqueIDType = Annotated[
    UniqueID, PlainValidator(lambda x: UniqueID(x)), PlainSerializer(lambda x: str(x))
]


class EnvVarStr(str):
    """String type that expands environment variables on validation.

    Automatically expands $VAR and ${VAR} patterns when validated.
    """

    @classmethod
    def __get_validators__(cls):
        """Yield validators for Pydantic.

        Yields:
            The validate method.
        """
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """Validate and expand environment variables.

        Args:
            v: The value to validate.

        Returns:
            The expanded string, or None.

        Raises:
            TypeError: If value is not a string.
        """
        if v is not None and not isinstance(v, str):
            raise TypeError("string required")
        return expandvars(v, "", False) if v else v

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"


class Env(BaseModel):
    """Environment configuration model.

    Attributes:
        env_type (EnvType): The environment type (dev, prod, etc.).
        label (Optional[str]): Optional label for the environment.
        account (Optional[str]): AWS account ID.
        region (Optional[str]): AWS region.
    """

    env_type: EnvType
    label: Annotated[str | None, PlainValidator(EnvVarStr.validate)] = None
    account: Annotated[str | None, PlainValidator(EnvVarStr.validate)] = None
    region: Annotated[str | None, PlainValidator(EnvVarStr.validate)] = None

    @property
    def env_name(self) -> str:
        """Get the environment name.

        Returns:
            The environment type value as a string.
        """
        return self.env_type.value

    @property
    def env_base(self) -> EnvBase:
        """Get the environment base identifier.

        Returns:
            EnvBase in format <env_name>[-<env_label>].
        """
        return EnvBase.from_type_and_label(self.env_type, self.label)

    @property
    def is_configured(self) -> bool:
        """Check if the environment has an AWS account configured.

        Returns:
            True if account is set, False otherwise.
        """
        return self.account is not None

    def to_env_var_map(self) -> MutableMapping[str, str]:
        """Convert to environment variable mapping.

        Returns:
            Dictionary of environment variable names to values.
        """
        return remove_null_values(
            dict(
                [
                    (ENV_TYPE_KEY, self.env_type),
                    (ENV_LABEL_KEY, self.label),
                    (ENV_BASE_KEY, self.env_base),
                ]
            ),  # type: ignore
            in_place=True,
        )


class CodePipelineBuildConfig(BaseModel):
    """Configuration for CodePipeline build settings.

    Attributes:
        ssh_key_secret_name (str): Name of the secret containing SSH key.
        docker_hub_credentials_secret_name (str): Name of the secret with
            Docker Hub credentials.
    """

    ssh_key_secret_name: str
    docker_hub_credentials_secret_name: str


class CodePipelineSourceConfig(BaseModel):
    """Configuration for CodePipeline source settings.

    Attributes:
        repository (str): The source repository.
        branch (str): The branch to build from.
        codestar_connection (Optional[UniqueIDType]): CodeStar connection ARN.
        oauth_secret_name (Optional[str]): Name of OAuth secret.

    Note:
        Either codestar_connection or oauth_secret_name must be set.
    """

    repository: str
    branch: Annotated[str, PlainValidator(EnvVarStr.validate)]
    codestar_connection: UniqueIDType | None = None
    oauth_secret_name: str | None = None

    @model_validator(mode="after")
    @classmethod
    def check_source_config(cls, v):
        if not v.codestar_connection and not v.oauth_secret_name:
            raise ValueError("Either codestar_connection or oauth_secret_name must be set")
        if v.codestar_connection and v.oauth_secret_name:
            logging.warning("Only one of codestar_connection or oauth_secret_name can be set")
        return v


class CodePipelineNotificationsConfig(BaseModel):
    """Configuration for CodePipeline notifications.

    Attributes:
        slack_channel_configuration_arn (Optional[str]): ARN of Slack channel.
        notify_on_failure (bool): Send notification on failure.
        notify_on_success (bool): Send notification on success.
    """

    slack_channel_configuration_arn: str | None
    notify_on_failure: bool = False
    notify_on_success: bool = False

    @property
    def notify_on_any(self) -> bool:
        """Check if any notifications are enabled.

        Returns:
            True if either failure or success notifications are enabled.
        """
        return self.notify_on_failure or self.notify_on_success


class PipelineConfig(BaseModel):
    """Complete pipeline configuration.

    Attributes:
        enable (bool): Whether the pipeline is enabled.
        build (CodePipelineBuildConfig): Build configuration.
        source (CodePipelineSourceConfig): Source configuration.
        notifications (CodePipelineNotificationsConfig): Notification settings.
    """

    enable: bool
    build: CodePipelineBuildConfig
    source: CodePipelineSourceConfig
    notifications: CodePipelineNotificationsConfig


class GlobalConfig(BaseModel):
    """Global project configuration.

    Attributes:
        pipeline_name (str): Name of the pipeline.
        stage_promotions (Dict[EnvType, EnvType]): Stage promotion mappings.
    """

    pipeline_name: str
    stage_promotions: dict[EnvType, EnvType]


class StageConfig(BaseModel):
    """Configuration for a deployment stage.

    Attributes:
        env (Env): Environment configuration.
        pipeline (Optional[PipelineConfig]): Optional pipeline configuration.
    """

    env: Env
    pipeline: PipelineConfig | None = None


DEFAULT_CONFIG_PATH = "configuration/project.yaml"


G = TypeVar("G", bound=GlobalConfig)
S = TypeVar("S", bound=StageConfig)
P = TypeVar("P", bound="BaseProjectConfig")


class BaseProjectConfig(BaseModel, Generic[G, S]):
    """Base class for project configuration.

    Generic base class that can be subclassed with custom global and stage
    config types.

    Attributes:
        global_config (G): Global configuration settings.
        default_config (S): Default stage configuration.
        default_config_overrides (Dict[EnvType, dict]): Per-environment overrides.
    """

    global_config: G
    default_config: S
    default_config_overrides: dict[EnvType, dict]

    @classmethod
    def get_global_config_cls(cls) -> type[G]:
        """Get the global config class type.

        Returns:
            The GlobalConfig type used by this project config.
        """
        return cls.model_fields["global_config"].annotation  # type: ignore

    @classmethod
    def get_stage_config_cls(cls) -> type[S]:
        """Get the stage config class type.

        Returns:
            The StageConfig type used by this project config.
        """
        return cls.model_fields["default_config"].annotation  # type: ignore

    def get_stage_config(self, env_type: str | EnvType, env_label: str | None = None) -> S:
        """Get stage config with environment type overrides.

        Args:
            env_type (Union[str, EnvType]): The environment type for the stage config.
            env_label (Optional[str]): Optional environment label override.
                Defaults to None (no override).

        Returns:
            A stage config object with applied overrides.

        Raises:
            Exception: If stage config model validation fails.
        """
        try:
            stage_config = self.get_stage_config_cls().model_validate(
                {
                    **DeepChainMap(
                        self.default_config_overrides[EnvType(env_type)],
                        self.default_config.model_dump(mode="json", exclude_unset=True),
                    ),
                }
            )
        except Exception as e:
            raise e

        if env_label is None:
            return stage_config
        else:
            stage_config.env.label = env_label
            return stage_config

    @classmethod
    def parse_file(cls: type[Self], path: str | Path, **kwargs) -> Self:  # type: ignore[override]
        """Parse configuration from a file.

        Args:
            path (Union[str, Path]): Path to the configuration file.
                Supports YAML and JSON formats.

        Returns:
            The parsed project configuration.
        """
        path = Path(path)

        if path.suffix in (".yml", ".yaml"):
            with open(path) as f:
                return cls.model_validate(yaml.safe_load(f))
        return cls.model_validate_json(json_data=path.read_text())

    @classmethod
    def load_config(cls: type[P], path: str | Path | None = None) -> P:
        """Load configuration from a file.

        Args:
            path (Optional[Union[str, Path]]): Path to config file.
                If None, searches for project.yaml in current directory.

        Returns:
            The loaded project configuration.

        Raises:
            AssertionError: If multiple or no project.yaml files found.
        """
        if path is None:
            paths = find_paths(
                Path.cwd(), include_dirs=False, include_files=True, includes=[r".*/project.yaml"]
            )
            assert len(paths) == 1, (
                f"Expected to find exactly one project.yaml file, but found {len(paths)}: {paths}"
            )
            path = paths[0]
        return cls.parse_file(path=path)

    @classmethod
    def load_stage_config(
        cls: type["BaseProjectConfig[G, S]"],
        env_type: str | EnvType,
        path: str | Path | None = None,
    ) -> S:
        """Load stage configuration for an environment type.

        Args:
            env_type (Union[str, EnvType]): The environment type.
            path (Optional[Union[str, Path]]): Path to config file.

        Returns:
            The stage configuration for the specified environment.
        """
        proj_config = cls.load_config(path)
        return proj_config.get_stage_config(env_type=env_type)


class ProjectConfig(BaseProjectConfig[GlobalConfig, StageConfig]):
    """Default project configuration using standard global and stage configs."""

    pass


class ConfigProvider:
    """Utility class for loading stage configurations."""

    @classmethod
    def get_stage_config(
        cls,
        env_type: str | EnvType,
        path: str | Path | None = None,
        project_config_cls: type[BaseProjectConfig[G, S]] = ProjectConfig,  # type: ignore[assignment]
    ) -> S:
        """Get stage configuration for an environment type.

        Args:
            env_type (Union[str, EnvType]): The environment type.
            path (Optional[Union[str, Path]]): Path to config file.
            project_config_cls (Type[BaseProjectConfig[G, S]]): Project config class
                to use. Defaults to ProjectConfig.

        Returns:
            The stage configuration.
        """
        proj_config = project_config_cls.load_config(path)
        return proj_config.get_stage_config(env_type=env_type)



from collections.abc import Iterable
from dataclasses import dataclass

import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_batch as batch
from aws_cdk import aws_ecr_assets as ecr_assets
from aws_cdk import aws_efs as efs
from aws_cdk import aws_events as events
from aws_cdk import aws_events_targets as events_targets
from aws_cdk import aws_s3 as s3
from aws_cdk import aws_stepfunctions as sfn

from aibs_informatics_cdk_lib.constructs_.efs.file_system import MountPointConfiguration
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.informatics.batch import (
    BatchInvokedBaseFragment,
    BatchInvokedLambdaFunction,
)


def get_data_path_stats_fragment(
    scope: constructs.Construct,
    id: str,
    env_base: EnvBase,
    aibs_informatics_docker_asset: ecr_assets.DockerImageAsset | str,
    batch_job_queue: batch.JobQueue | str,
    scaffolding_bucket: s3.Bucket,
    mount_point_configs: Iterable[MountPointConfiguration] | None = None,
    memory: int = 1024,
    vcpus: int = 1,
) -> BatchInvokedLambdaFunction:
    """Returns a BatchInvokedLambdaFunction fragment for getting data path stats of EFS/S3 path

    Args:
        scope (constructs.Construct): scope
        id (str): id of the fragment
        env_base (EnvBase): env base
        aibs_informatics_docker_asset (Union[ecr_assets.DockerImageAsset, str]): docker image asset or image uri
            that has the get_data_path_stats_handler function
        batch_job_queue (Union[batch.JobQueue, str]): default batch job queue or job queue name str that
            the batch job will be submitted to. This can be override by the payload.
        scaffolding_bucket (s3.Bucket): primary bucket used for request/response json blobs used in
        mount_point_configs (Optional[Iterable[MountPointConfiguration]], optional): Default EFS volumes to mount.
            Defaults to None.
        memory (int, optional): memory needed. Defaults to 1024.
        vcpus (int, optional): vcpus needed. Defaults to 1.

    Returns:
        BatchInvokedLambdaFunction fragment for getting data path stats
    """  # noqa: E501
    fragment = BatchInvokedLambdaFunction(
        scope=scope,
        id=id,
        env_base=env_base,
        name="get-data-path-stats",
        image=(
            aibs_informatics_docker_asset
            if isinstance(aibs_informatics_docker_asset, str)
            else aibs_informatics_docker_asset.image_uri
        ),
        handler="aibs_informatics_aws_lambda.handlers.data_sync.get_data_path_stats_handler",
        job_queue=(
            batch_job_queue if isinstance(batch_job_queue, str) else batch_job_queue.job_queue_name
        ),
        bucket_name=scaffolding_bucket.bucket_name,
        memory=memory,
        vcpus=vcpus,
        mount_point_configs=list(mount_point_configs) if mount_point_configs else None,
    )
    return fragment


def outdated_data_path_scanner_fragment(
    scope: constructs.Construct,
    id: str,
    env_base: EnvBase,
    aibs_informatics_docker_asset: ecr_assets.DockerImageAsset | str,
    batch_job_queue: batch.JobQueue | str,
    scaffolding_bucket: s3.Bucket,
    mount_point_configs: Iterable[MountPointConfiguration] | None = None,
    memory: int = 1024,
    vcpus: int = 1,
) -> BatchInvokedLambdaFunction:
    """Returns a BatchInvokedLambdaFunction fragment for scanning outdated data paths of EFS/S3 path root

    Args:
        scope (constructs.Construct): scope
        id (str): id of the fragment
        env_base (EnvBase): env base
        aibs_informatics_docker_asset (Union[ecr_assets.DockerImageAsset, str]): docker image asset or image uri
            that has the outdated_data_path_scanner_handler function
        batch_job_queue (Union[batch.JobQueue, str]): default batch job queue or job queue name str that
            the batch job will be submitted to. This can be override by the payload.
        scaffolding_bucket (s3.Bucket): primary bucket used for request/response json blobs used in
        mount_point_configs (Optional[Iterable[MountPointConfiguration]], optional): Default EFS volumes to mount.
            Defaults to None.
        memory (int, optional): memory needed. Defaults to 1024.
        vcpus (int, optional): vcpus needed. Defaults to 1.

    Returns:
        BatchInvokedLambdaFunction fragment for scanning outdated data paths
    """  # noqa: E501

    fragment = BatchInvokedLambdaFunction(
        scope=scope,
        id=id,
        env_base=env_base,
        name="outdated-data-path-scanner",
        image=(
            aibs_informatics_docker_asset
            if isinstance(aibs_informatics_docker_asset, str)
            else aibs_informatics_docker_asset.image_uri
        ),
        handler="aibs_informatics_aws_lambda.handlers.data_sync.outdated_data_path_scanner_handler",
        job_queue=(
            batch_job_queue if isinstance(batch_job_queue, str) else batch_job_queue.job_queue_name
        ),
        bucket_name=scaffolding_bucket.bucket_name,
        memory=memory,
        vcpus=vcpus,
        # mount_points=mount_points,
        # volumes=volumes,
        mount_point_configs=list(mount_point_configs) if mount_point_configs else None,
    )
    return fragment


def remove_data_paths_fragment(
    scope: constructs.Construct,
    id: str,
    env_base: EnvBase,
    aibs_informatics_docker_asset: ecr_assets.DockerImageAsset | str,
    batch_job_queue: batch.JobQueue | str,
    scaffolding_bucket: s3.Bucket,
    mount_point_configs: Iterable[MountPointConfiguration] | None = None,
    memory: int = 1024,
    vcpus: int = 1,
) -> BatchInvokedLambdaFunction:
    """Returns a BatchInvokedLambdaFunction fragment for removing data paths (EFS / S3) during execution of a Step Function

    Args:
        scope (constructs.Construct): scope
        id (str): id of the fragment
        env_base (EnvBase): env base
        aibs_informatics_docker_asset (Union[ecr_assets.DockerImageAsset, str]): docker image asset or image uri
            that has the remove_data_paths_handler function
        batch_job_queue (Union[batch.JobQueue, str]): default batch job queue or job queue name str that
            the batch job will be submitted to. This can be override by the payload.
        scaffolding_bucket (s3.Bucket): primary bucket used for request/response json blobs used in
        mount_point_configs (Optional[Iterable[MountPointConfiguration]], optional): Default EFS volumes to mount.
            Defaults to None.
        memory (int, optional): memory needed. Defaults to 1024.
        vcpus (int, optional): vcpus needed. Defaults to 1.

    Returns:
        BatchInvokedLambdaFunction fragment for removing data paths
    """  # noqa: E501
    fragment = BatchInvokedLambdaFunction(
        scope=scope,
        id=id,
        env_base=env_base,
        name="remove-data-paths",
        image=(
            aibs_informatics_docker_asset
            if isinstance(aibs_informatics_docker_asset, str)
            else aibs_informatics_docker_asset.image_uri
        ),
        handler="aibs_informatics_aws_lambda.handlers.data_sync.remove_data_paths_handler",
        job_queue=(
            batch_job_queue if isinstance(batch_job_queue, str) else batch_job_queue.job_queue_name
        ),
        bucket_name=scaffolding_bucket.bucket_name,
        memory=memory,
        vcpus=vcpus,
        mount_point_configs=list(mount_point_configs) if mount_point_configs else None,
    )
    return fragment


@dataclass
class CleanFileSystemTriggerConfig:
    file_system: efs.FileSystem
    path: str
    days_since_last_accessed: float = 3.0
    max_depth: int | None = None
    min_depth: int | None = None
    min_size_bytes_allowed: int = 0

    schedule: events.Schedule = events.Schedule.cron(minute="0", hour="9")

    def to_dict(self):
        d = {
            "path": f"{self.file_system.file_system_id}:{self.path}",
            "days_since_last_accessed": self.days_since_last_accessed,
        }
        if self.max_depth is not None:
            d["max_depth"] = self.max_depth
        if self.min_depth is not None:
            d["min_depth"] = self.min_depth
        if self.min_size_bytes_allowed > 0:
            d["min_size_bytes_allowed"] = self.min_size_bytes_allowed
        return d


@dataclass
class CleanFileSystemTriggerRuleConfig:
    rule_name: str
    trigger_configs: list[CleanFileSystemTriggerConfig]
    # https://docs.aws.amazon.com/eventbridge/latest/userguide/eb-create-rule-schedule.html#cron-expressions
    # Want to run around 00:00 in PST by default
    schedule: events.Schedule = events.Schedule.cron(minute="0", hour="9")

    @property
    def description(self):
        return (
            "Daily trigger for EFS file cleanup (older than "
            f"{sorted({_.days_since_last_accessed for _ in self.trigger_configs})} "
            "days) for target subdirectories"
        )

    def create_rule(
        self,
        scope: constructs.Construct,
        clean_file_system_state_machine: sfn.StateMachine,
    ):
        return events.Rule(
            scope,
            self.rule_name,
            rule_name=self.rule_name,
            description=self.description,
            enabled=True,
            schedule=self.schedule,
            targets=[
                events_targets.SfnStateMachine(
                    clean_file_system_state_machine,
                    input=events.RuleTargetInput.from_object(config.to_dict()),
                )
                for config in self.trigger_configs
            ],  # type: ignore[arg-type]  # jsii implementation issue - https://github.com/aws/jsii/issues/4531
        )


class CleanFileSystemFragment(BatchInvokedBaseFragment):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        aibs_informatics_docker_asset: ecr_assets.DockerImageAsset | str,
        batch_job_queue: batch.JobQueue | str,
        scaffolding_bucket: s3.Bucket,
        mount_point_configs: Iterable[MountPointConfiguration] | None = None,
        memory: int = 1024,
        vcpus: int = 1,
    ) -> None:
        """Clean up the file system by scanning for outdated data paths and removing them

        Args:
            scope (Construct): construct scope
            id (str): id
            env_base (EnvBase): env base
            aibs_informatics_docker_asset (DockerImageAsset|str): Docker image asset or image uri
                str for the aibs informatics aws lambda
            batch_job_queue (JobQueue|str): Default batch job queue or job queue name str that
                the batch job will be submitted to. This can be override by the payload.
            primary_bucket (Bucket): Primary bucket used for request/response json blobs used in
                the batch invoked lambda function.
            mount_point_configs (Optional[Iterable[MountPointConfiguration]], optional):
                List of mount point configurations to use. These can be overridden in the payload.
            memory (int, optional): memory needed. Defaults to 1024.
                This memory value is used for both the outdated path scanner and removal of data paths.
            vcpus (int, optional): vcpus needed. Defaults to 1.
                This memory value is used for both the outdated path scanner and removal of data paths.
        """  # noqa: E501
        super().__init__(scope, id, env_base)

        aibs_informatics_image_uri = (
            aibs_informatics_docker_asset
            if isinstance(aibs_informatics_docker_asset, str)
            else aibs_informatics_docker_asset.image_uri
        )

        start_pass_state = sfn.Pass(
            self,
            "Data Cleanup: Start",
        )

        self.outdated_data_path_scanner = outdated_data_path_scanner_fragment(
            self,
            "Scan for Outdated Data Paths",
            env_base=self.env_base,
            aibs_informatics_docker_asset=aibs_informatics_image_uri,
            batch_job_queue=batch_job_queue,
            scaffolding_bucket=scaffolding_bucket,
            mount_point_configs=mount_point_configs,
            memory=memory,
            vcpus=vcpus,
        )

        self.remove_data_paths = remove_data_paths_fragment(
            self,
            "Remove Data Paths",
            env_base=self.env_base,
            aibs_informatics_docker_asset=aibs_informatics_image_uri,
            batch_job_queue=batch_job_queue,
            scaffolding_bucket=scaffolding_bucket,
            mount_point_configs=mount_point_configs,
            memory=memory,
            vcpus=vcpus,
        )

        # fmt: off
        self.definition = (
            start_pass_state
            .next(self.outdated_data_path_scanner.enclose())
            .next(self.remove_data_paths.enclose())
        )
        # fmt: on



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, cast

import constructs
from aibs_informatics_core.env import EnvBase
from aibs_informatics_core.utils.tools.dicttools import convert_key_case
from aibs_informatics_core.utils.tools.strtools import pascalcase
from aws_cdk import JsonNull
from aws_cdk import aws_stepfunctions as sfn

from aibs_informatics_cdk_lib.constructs_.efs.file_system import MountPointConfiguration
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.base import EnvBaseStateMachineFragment
from aibs_informatics_cdk_lib.constructs_.sfn.states.batch import BatchOperation
from aibs_informatics_cdk_lib.constructs_.sfn.states.common import CommonOperation

if TYPE_CHECKING:
    from mypy_boto3_batch.type_defs import MountPointTypeDef, VolumeTypeDef
else:  # pragma: no cover
    MountPointTypeDef = dict
    VolumeTypeDef = dict


class AWSBatchMixins:
    @classmethod
    def convert_to_mount_point_and_volumes(
        cls,
        mount_point_configs: list[MountPointConfiguration],
    ) -> tuple[list[MountPointTypeDef], list[VolumeTypeDef]]:
        mount_points = []
        volumes = []
        for i, mpc in enumerate(mount_point_configs):
            mount_points.append(
                convert_key_case(mpc.to_batch_mount_point(f"efs-vol{i}"), pascalcase)
            )
            volumes.append(convert_key_case(mpc.to_batch_volume(f"efs-vol{i}"), pascalcase))
        return cast(tuple[list[MountPointTypeDef], list[VolumeTypeDef]], (mount_points, volumes))


class SubmitJobFragment(EnvBaseStateMachineFragment, AWSBatchMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        name: str,
        job_queue: str,
        image: str,
        command: list[str] | str | None = None,
        environment: Mapping[str, str] | str | None = None,
        memory: int | str | None = None,
        vcpus: int | str | None = None,
        gpu: int | str | None = None,
        mount_points: list[MountPointTypeDef] | str | None = None,
        volumes: list[VolumeTypeDef] | str | None = None,
        platform_capabilities: list[Literal["EC2", "FARGATE"]] | str | None = None,
        job_role_arn: str | None = None,
    ) -> None:
        super().__init__(scope, id, env_base)

        register_chain = BatchOperation.register_job_definition(
            self,
            id,
            command=command,
            image=image,
            job_definition_name=name,
            memory=memory,
            vcpus=vcpus,
            gpu=gpu,
            mount_points=mount_points,
            volumes=volumes,
            platform_capabilities=platform_capabilities,
            job_role_arn=job_role_arn,
        )

        submit_chain = BatchOperation.submit_job(
            self,
            id,
            job_name=name,
            job_definition=sfn.JsonPath.string_at("$.taskResult.register.JobDefinitionArn"),
            job_queue=job_queue,
            command=command,
            environment=environment,
            vcpus=vcpus,
            memory=memory,
            gpu=gpu,
        )

        deregister_chain = BatchOperation.deregister_job_definition(
            self,
            id,
            job_definition=sfn.JsonPath.string_at("$.taskResult.register.JobDefinitionArn"),
        )

        try_catch_deregister_chain = BatchOperation.deregister_job_definition(
            self,
            id + " FAIL",
            job_definition=sfn.JsonPath.string_at("$.taskResult.register.JobDefinitionArn"),
        )

        register = CommonOperation.enclose_chainable(
            self, id + " Register", register_chain, result_path="$.taskResult.register"
        )
        # submit = StateMachineFragment.enclose(
        submit = CommonOperation.enclose_chainable(
            self, id + " Submit", submit_chain, result_path="$.taskResult.submit"
        ).to_single_state(id=f"{id} Enclosure", output_path="$[0]")
        deregister = CommonOperation.enclose_chainable(
            self, id + " Deregister", deregister_chain, result_path="$.taskResult.deregister"
        )
        try_catch_deregister = CommonOperation.enclose_chainable(
            self,
            f"{id} Deregister FAIL",
            try_catch_deregister_chain,
            result_path="$.taskResult.deregister",
        )
        submit.add_catch(
            try_catch_deregister.next(
                sfn.Fail(
                    self,
                    id + " FAIL",
                    cause_path=sfn.JsonPath.string_at("$.taskResult.submit.Cause"),
                    error_path=sfn.JsonPath.string_at("$.taskResult.submit.Error"),
                )
            ),
            result_path="$.taskResult.submit",
            errors=["States.ALL"],
        )

        self.definition = register.next(submit).next(deregister)

    @classmethod
    def from_defaults(
        cls,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        name: str,
        job_queue: str,
        image: str,
        command: str = "",
        memory: str | int = 1024,
        vcpus: str | int = 1,
        gpu: str | int = 0,
        environment: Mapping[str, str] | None = None,
        mount_point_configs: list[MountPointConfiguration] | None = None,
        job_role_arn: str | None = None,
    ) -> "SubmitJobFragment":
        defaults: dict[str, Any] = {}
        defaults["command"] = command
        defaults["job_queue"] = job_queue
        defaults["environment"] = environment or {}
        defaults["image"] = image
        defaults["memory"] = str(memory)
        defaults["vcpus"] = str(vcpus)
        defaults["gpu"] = str(gpu)
        defaults["platform_capabilities"] = ["EC2"]
        defaults["job_role_arn"] = job_role_arn or JsonNull.INSTANCE

        if mount_point_configs:
            mount_points, volumes = cls.convert_to_mount_point_and_volumes(mount_point_configs)
            defaults["mount_points"] = mount_points
            defaults["volumes"] = volumes

        submit_job = SubmitJobFragment(
            scope,
            id,
            env_base=env_base,
            name=name,
            image=sfn.JsonPath.string_at("$.request.image"),
            command=sfn.JsonPath.string_at("$.request.command"),
            job_queue=sfn.JsonPath.string_at("$.request.job_queue"),
            environment=sfn.JsonPath.string_at("$.request.environment"),
            memory=sfn.JsonPath.string_at("$.request.memory"),
            vcpus=sfn.JsonPath.string_at("$.request.vcpus"),
            gpu=sfn.JsonPath.string_at("$.request.gpu"),
            mount_points=sfn.JsonPath.string_at("$.request.mount_points"),
            volumes=sfn.JsonPath.string_at("$.request.volumes"),
            platform_capabilities=sfn.JsonPath.string_at("$.request.platform_capabilities"),
            job_role_arn=sfn.JsonPath.string_at("$.request.job_role_arn"),
        )

        # Now we need to add the start and merge states and add to the definition
        start = sfn.Pass(
            submit_job,
            "Start",
            parameters={
                "input": sfn.JsonPath.string_at("$"),
                "default": defaults,
            },
        )
        merge = sfn.Pass(
            submit_job,
            "Merge",
            parameters={
                "request": sfn.JsonPath.json_merge(
                    sfn.JsonPath.object_at("$.default"), sfn.JsonPath.object_at("$.input")
                ),
            },
        )

        submit_job.definition = start.next(merge).next(submit_job.definition)
        return submit_job



from aws_cdk import aws_batch as batch

from aibs_informatics_cdk_lib.constructs_.batch.infrastructure import BatchEnvironmentConfig
from aibs_informatics_cdk_lib.constructs_.batch.instance_types import (
    LAMBDA_LARGE_INSTANCE_TYPES,
    LAMBDA_MEDIUM_INSTANCE_TYPES,
    LAMBDA_SMALL_INSTANCE_TYPES,
    ON_DEMAND_INSTANCE_TYPES,
    SPOT_INSTANCE_TYPES,
    TRANSFER_INSTANCE_TYPES,
)

LOW_PRIORITY_BATCH_ENV_CONFIG = BatchEnvironmentConfig(
    allocation_strategy=batch.AllocationStrategy.SPOT_PRICE_CAPACITY_OPTIMIZED,
    instance_types=[*SPOT_INSTANCE_TYPES],
    use_spot=True,
    use_fargate=False,
    use_public_subnets=False,
)
NORMAL_PRIORITY_BATCH_ENV_CONFIG = BatchEnvironmentConfig(
    allocation_strategy=batch.AllocationStrategy.SPOT_PRICE_CAPACITY_OPTIMIZED,
    instance_types=[*SPOT_INSTANCE_TYPES],
    use_spot=True,
    use_fargate=False,
    use_public_subnets=False,
)
HIGH_PRIORITY_BATCH_ENV_CONFIG = BatchEnvironmentConfig(
    allocation_strategy=batch.AllocationStrategy.BEST_FIT,
    instance_types=[*ON_DEMAND_INSTANCE_TYPES],
    use_spot=False,
    use_fargate=False,
    use_public_subnets=False,
)
PUBLIC_SUBNET_BATCH_ENV_CONFIG = BatchEnvironmentConfig(
    allocation_strategy=batch.AllocationStrategy.BEST_FIT,
    instance_types=[*TRANSFER_INSTANCE_TYPES],
    use_spot=False,
    use_fargate=False,
    use_public_subnets=True,
)

LAMBDA_BATCH_ENV_CONFIG = BatchEnvironmentConfig(
    allocation_strategy=batch.AllocationStrategy.BEST_FIT,
    instance_types=[
        *LAMBDA_SMALL_INSTANCE_TYPES,
        *LAMBDA_MEDIUM_INSTANCE_TYPES,
        *LAMBDA_LARGE_INSTANCE_TYPES,
    ],
    use_spot=False,
    use_fargate=False,
    use_public_subnets=False,
)
LAMBDA_SMALL_BATCH_ENV_CONFIG = BatchEnvironmentConfig(
    allocation_strategy=batch.AllocationStrategy.BEST_FIT,
    instance_types=[*LAMBDA_SMALL_INSTANCE_TYPES],
    use_spot=False,
    use_fargate=False,
    use_public_subnets=False,
)
LAMBDA_MEDIUM_BATCH_ENV_CONFIG = BatchEnvironmentConfig(
    allocation_strategy=batch.AllocationStrategy.BEST_FIT,
    instance_types=[*LAMBDA_MEDIUM_INSTANCE_TYPES],
    use_spot=False,
    use_fargate=False,
    use_public_subnets=False,
)
LAMBDA_LARGE_BATCH_ENV_CONFIG = BatchEnvironmentConfig(
    allocation_strategy=batch.AllocationStrategy.BEST_FIT,
    instance_types=[*LAMBDA_LARGE_INSTANCE_TYPES],
    use_spot=False,
    use_fargate=False,
    use_public_subnets=False,
)



from typing import cast

import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_lambda as lambda_
from aws_cdk import aws_stepfunctions as sfn
from aws_cdk import aws_stepfunctions_tasks as stepfn_tasks

from aibs_informatics_cdk_lib.constructs_.sfn.fragments.base import EnvBaseStateMachineFragment


class LambdaFunctionFragment(EnvBaseStateMachineFragment):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        lambda_function: lambda_.Function,
    ) -> None:
        """Creates a single task state machine fragment for a generic lambda function

                    [Pass]          # Start State. Augments input.
                      ||            #
                  [lambda fn]
               (x)  //   \\ (o)     #
                [Fail]    ||        # Catch state for failed lambda execution
                          ||        #
                        [Pass]      # End state. Reforms output of lambda as sfn output

        Returns:
            The state machine fragment
        """
        super().__init__(scope, id, env_base)

        lambda_task = stepfn_tasks.LambdaInvoke(
            self,
            f"{lambda_function.function_name} Function Execution",
            lambda_function=cast(lambda_.IFunction, lambda_function),
            payload_response_only=True,
        )

        self.definition = sfn.Chain.start(lambda_task)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal

import constructs
from aibs_informatics_aws_utils.constants.lambda_ import (
    AWS_LAMBDA_EVENT_PAYLOAD_KEY,
    AWS_LAMBDA_EVENT_RESPONSE_LOCATION_KEY,
    AWS_LAMBDA_FUNCTION_HANDLER_KEY,
    AWS_LAMBDA_FUNCTION_NAME_KEY,
)
from aibs_informatics_aws_utils.constants.s3 import S3_SCRATCH_KEY_PREFIX
from aibs_informatics_core.env import EnvBase
from aws_cdk import JsonNull
from aws_cdk import aws_iam as iam
from aws_cdk import aws_stepfunctions as sfn

from aibs_informatics_cdk_lib.common.aws.iam_utils import (
    batch_policy_statement,
    s3_policy_statement,
)
from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins
from aibs_informatics_cdk_lib.constructs_.efs.file_system import MountPointConfiguration
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.base import EnvBaseStateMachineFragment
from aibs_informatics_cdk_lib.constructs_.sfn.fragments.batch import (
    AWSBatchMixins,
    SubmitJobFragment,
)
from aibs_informatics_cdk_lib.constructs_.sfn.states.common import CommonOperation
from aibs_informatics_cdk_lib.constructs_.sfn.states.s3 import S3Operation

if TYPE_CHECKING:
    from mypy_boto3_batch.type_defs import MountPointTypeDef, VolumeTypeDef
else:
    MountPointTypeDef = dict
    VolumeTypeDef = dict


class BatchInvokedBaseFragment(EnvBaseStateMachineFragment, EnvBaseConstructMixins):
    @property
    def required_inline_policy_statements(self) -> list[iam.PolicyStatement]:
        return [
            *super().required_inline_policy_statements,
            batch_policy_statement(self.env_base),
            s3_policy_statement(self.env_base),
            iam.PolicyStatement(
                sid="PassRoleForBatchJobs",
                actions=["iam:PassRole"],
                effect=iam.Effect.ALLOW,
                resources=["*"],
                conditions={"StringLike": {"iam:PassedToService": "ecs-tasks.amazonaws.com"}},
            ),
        ]


class BatchInvokedLambdaFunction(BatchInvokedBaseFragment, AWSBatchMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        name: str,
        image: str,
        handler: str,
        job_queue: str,
        bucket_name: str,
        key_prefix: str | None = None,
        payload_path: str | None = None,
        command: list[str] | str | None = None,
        environment: Mapping[str, str] | None = None,
        memory: int | str | None = None,
        vcpus: int | str | None = None,
        gpu: int | str | None = None,
        mount_points: list[MountPointTypeDef] | str | None = None,
        volumes: list[VolumeTypeDef] | str | None = None,
        mount_point_configs: list[MountPointConfiguration] | None = None,
        platform_capabilities: list[Literal["EC2", "FARGATE"]] | str | None = None,
        job_role_arn: str | None = None,
    ) -> None:
        """Invoke a command on image via batch with a payload from s3

        This fragment creates a state machine fragment that:
            1. Puts a payload to s3
            2. Submits a batch job
            3. Gets the response from s3

        The payload is written to s3://<bucket_name>/<key_prefix>/<execution_name>/request.json
        The response is read from s3://<bucket_name>/<key_prefix>/<execution_name>/response.json

        The batch job will be fed the following environment variables:
            - AWS_LAMBDA_FUNCTION_NAME: name of lambda function
            - AWS_LAMBDA_FUNCTION_HANDLER: handler of lambda function
            - AWS_LAMBDA_EVENT_PAYLOAD: The s3 location of the event payload
            - AWS_LAMBDA_EVENT_RESPONSE_LOCATION: The s3 location to write the response to

        IMPORTANT:
            - Batch job queue / compute environment must have permissions to read/write to the bucket.


        Args:
            scope (Construct): construct scope
            id (str): id
            env_base (EnvBase): env base
            name (str): Name of the lambda function. This can be a reference path (e.g. "$.name")
            image (str): Image URI or name. This can be a reference path (e.g. "$.image")
            handler (str): handler of lambda function. This should describe a fully qualified path to function handler. This can be a reference path (e.g. "$.handler")
            job_queue (str): Job queue to submit job to. This can be a reference path (e.g. "$.job_queue")
            bucket_name (str): S3 Bucket name to write payload to and read response from. This can be a reference path (e.g. "$.bucket_name")
            key_prefix (str | None): Key prefix to write payload to and read response from. If not provided, `scratch/` is used. Can be a reference path (e.g. "$.key_prefix")
            payload_path (str | None): Optionally specify the reference path of the event payload. Defaults to "$".
            command (List[str] | str | None): Command to run in container. Can be a reference path (e.g. "$.command"). If unspecified, the container's CMD is used.
            environment (Mapping[str, str] | None): Additional environment variables to specify. These are added to default environment variables.
            memory (int | str | None): Memory in MiB (either int or reference path str). Defaults to None.
            vcpus (int | str | None): Number of vCPUs (either int or reference path str). Defaults to None.
            gpu (int | str | None): Number of GPUs (either int or reference path str). Defaults to None.
            mount_points (List[MountPointTypeDef] | None): List of mount points to add to state machine. Defaults to None.
            volumes (List[VolumeTypeDef] | None): List of volumes to add to state machine. Defaults to None.
            platform_capabilities (List[Literal["EC2", "FARGATE"]] | str | None): platform capabilities to use. This can be a reference path (e.g. "$.platform_capabilities")
            job_role_arn (str | None): Job role arn to use for the job. This can be a reference path (e.g. "$.job_role_arn")
        """  # noqa: E501
        super().__init__(scope, id, env_base)
        key_prefix = key_prefix or S3_SCRATCH_KEY_PREFIX

        request_key = sfn.JsonPath.format(
            f"{key_prefix}{{}}/{{}}/request.json",
            sfn.JsonPath.execution_name,
            sfn.JsonPath.string_at("$.taskResult.prep.task_id"),
        )
        response_key = sfn.JsonPath.format(
            f"{key_prefix}{{}}/{{}}/response.json",
            sfn.JsonPath.execution_name,
            sfn.JsonPath.string_at("$.taskResult.prep.task_id"),
        )

        start = sfn.Pass(
            self,
            f"{id} Prep S3 Keys",
            parameters={
                "task_id": sfn.JsonPath.uuid(),
            },
            result_path="$.taskResult.prep",
        )

        if mount_point_configs:
            if mount_points or volumes:
                raise ValueError("Cannot specify both mount_point_configs and mount_points")
            mount_points, volumes = self.convert_to_mount_point_and_volumes(mount_point_configs)

        put_payload = S3Operation.put_payload(
            self,
            f"{id} Put Request to S3",
            payload=payload_path or sfn.JsonPath.entire_payload,
            bucket_name=bucket_name,
            key=request_key,
            result_path="$.taskResult.put",
        )

        default_environment = {
            AWS_LAMBDA_FUNCTION_NAME_KEY: name,
            AWS_LAMBDA_FUNCTION_HANDLER_KEY: handler,
            AWS_LAMBDA_EVENT_PAYLOAD_KEY: sfn.JsonPath.format(
                "s3://{}/{}",
                sfn.JsonPath.string_at("$.taskResult.put.Bucket"),
                sfn.JsonPath.string_at("$.taskResult.put.Key"),
            ),
            AWS_LAMBDA_EVENT_RESPONSE_LOCATION_KEY: sfn.JsonPath.format(
                "s3://{}/{}",
                bucket_name,
                response_key,
            ),
            EnvBase.ENV_BASE_KEY: self.env_base,
            "AWS_REGION": self.aws_region,
            "AWS_ACCOUNT_ID": self.aws_account,
        }

        submit_job = SubmitJobFragment(
            self,
            f"{id} Batch",
            env_base=env_base,
            name=name,
            job_queue=job_queue,
            command=command or [],
            image=image,
            environment={
                **(environment if environment else {}),
                **default_environment,
            },
            memory=memory,
            vcpus=vcpus,
            gpu=gpu,
            mount_points=mount_points or [],
            volumes=volumes or [],
            platform_capabilities=platform_capabilities,
            job_role_arn=job_role_arn,
        )

        get_response = S3Operation.get_payload(
            self,
            f"{id}",
            bucket_name=bucket_name,
            key=response_key,
        ).to_single_state(
            f"{id} Get Response from S3",
            output_path="$[0]",
        )

        self.definition = start.next(put_payload).next(submit_job).next(get_response)

    @property
    def start_state(self) -> sfn.State:
        return self.definition.start_state

    @property
    def end_states(self) -> list[sfn.INextable]:
        return self.definition.end_states

    @classmethod
    def with_defaults(
        cls,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        name: str,
        job_queue: str,
        bucket_name: str,
        key_prefix: str | None = None,
        image_path: str | None = None,
        handler_path: str | None = None,
        payload_path: str | None = None,
        overrides_path: str | None = None,
        command: list[str] | None = None,
        memory: int | str = "1024",
        vcpus: int | str = "1",
        gpu: int | str = "0",
        environment: Mapping[str, str] | None = None,
        mount_point_configs: list[MountPointConfiguration] | None = None,
        platform_capabilities: list[Literal["EC2", "FARGATE"]] | None = None,
        job_role_arn: iam.Role | str | None = None,
    ) -> "BatchInvokedLambdaFunction":
        defaults: dict[str, Any] = {}

        defaults["job_queue"] = job_queue
        defaults["memory"] = str(memory)
        defaults["vcpus"] = str(vcpus)
        defaults["gpu"] = str(gpu)
        defaults["environment"] = environment or {}
        defaults["platform_capabilities"] = platform_capabilities or ["EC2"]
        defaults["bucket_name"] = bucket_name
        defaults["job_role_arn"] = job_role_arn or JsonNull.INSTANCE

        defaults["command"] = command if command else []

        if mount_point_configs:
            mount_points, volumes = cls.convert_to_mount_point_and_volumes(mount_point_configs)
            defaults["mount_points"] = mount_points
            defaults["volumes"] = volumes

        fragment = BatchInvokedLambdaFunction(
            scope,
            id,
            env_base=env_base,
            name=name,
            image=sfn.JsonPath.string_at("$.image"),
            handler=sfn.JsonPath.string_at("$.handler"),
            job_queue=sfn.JsonPath.string_at("$.merged.job_queue"),
            bucket_name=sfn.JsonPath.string_at("$.merged.bucket_name"),
            key_prefix=key_prefix,
            payload_path=sfn.JsonPath.string_at("$.payload"),
            command=sfn.JsonPath.string_at("$.merged.command"),
            environment=environment,
            memory=sfn.JsonPath.string_at("$.merged.memory"),
            vcpus=sfn.JsonPath.string_at("$.merged.vcpus"),
            # TODO: Handle GPU parameter better - right now, we cannot handle cases where it is
            # not specified. Setting to zero causes issues with the Batch API.
            # If it is set to zero, then the json list of resources are not properly set.
            gpu=sfn.JsonPath.string_at("$.merged.gpu"),
            mount_points=sfn.JsonPath.string_at("$.merged.mount_points"),
            volumes=sfn.JsonPath.string_at("$.merged.volumes"),
            platform_capabilities=sfn.JsonPath.string_at("$.merged.platform_capabilities"),
            job_role_arn=sfn.JsonPath.string_at("$.merged.job_role_arn"),
        )

        start = sfn.Pass(
            fragment,
            f"Start {id}",
            parameters={
                "image": sfn.JsonPath.string_at(image_path or "$.image"),
                "handler": sfn.JsonPath.string_at(handler_path or "$.handler"),
                "payload": sfn.JsonPath.object_at(payload_path or "$.payload"),
                # We will merge the rest with the defaults
                "input": sfn.JsonPath.object_at(overrides_path if overrides_path else "$"),
            },
        )

        merge_chain = CommonOperation.merge_defaults(
            fragment,
            f"Merge {id}",
            input_path="$.input",
            defaults=defaults,
            result_path="$.merged",
        )

        fragment.definition = start.next(merge_chain).next(fragment.definition)
        return fragment


class BatchInvokedExecutorFragment(BatchInvokedBaseFragment, AWSBatchMixins):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        name: str,
        image: str,
        executor: str,
        job_queue: str,
        bucket_name: str,
        key_prefix: str | None = None,
        payload_path: str | None = None,
        environment: Mapping[str, str] | str | None = None,
        memory: int | str | None = None,
        vcpus: int | str | None = None,
        mount_point_configs: list[MountPointConfiguration] | None = None,
        mount_points: list[MountPointTypeDef] | None = None,
        volumes: list[VolumeTypeDef] | None = None,
        platform_capabilities: list[Literal["EC2", "FARGATE"]] | str | None = None,
        job_role_arn: str | None = None,
    ) -> None:
        """Invoke an executor in an image via batch with a payload from s3

        This targets any subclassing of `aibs_informatics_core.executors.base.ExecutorBase`
        - https://github.com/AllenInstitute/aibs-informatics-core/blob/main/src/aibs_informatics_core/executors/base.py


        This fragment creates a state machine fragment that:
            1. Puts a payload to s3
            2. Submits a batch job
            3. Gets the response from s3

        The payload is written to s3://<bucket_name>/<key_prefix>/<execution_name>/request.json
        The response is read from s3://<bucket_name>/<key_prefix>/<execution_name>/response.json

        IMPORTANT:
            - Batch job queue / compute environment must have permissions to read/write to the bucket.

        Args:
            scope (Construct): construct scope
            id (str): id
            env_base (EnvBase): env base
            name (str): Name of the lambda function. This can be a reference path (e.g. "$.name")
            image (str): Image URI or name. This can be a reference path (e.g. "$.image")
            executor (str): qualified name of executor class. This should describe a fully qualified path to function handler. This can be a reference path (e.g. "$.handler")
            job_queue (str): Job queue to submit job to. This can be a reference path (e.g. "$.job_queue")
            bucket_name (str): S3 Bucket name to write payload to and read response from. This can be a reference path (e.g. "$.bucket_name")
            key_prefix (str | None): Key prefix to write payload to and read response from. If not provided, `scratch/` is used. Can be a reference path (e.g. "$.key_prefix")
            payload_path (str | None): Optionally specify the reference path of the event payload. Defaults to "$".
            command (List[str] | str | None): Command to run in container. Can be a reference path (e.g. "$.command"). If unspecified, the container's CMD is used.
            environment (Mapping[str, str] | str | None): environment variables to specify. This can be a reference path (e.g. "$.environment")
            memory (int | str | None): Memory in MiB (either int or reference path str). Defaults to None.
            vcpus (int | str | None): Number of vCPUs (either int or reference path str). Defaults to None.
            mount_points (List[MountPointTypeDef] | None): List of mount points to add to state machine. Defaults to None.
            volumes (List[VolumeTypeDef] | None): List of volumes to add to state machine. Defaults to None.
            platform_capabilities (List[Literal["EC2", "FARGATE"]] | str | None): platform capabilities to use. This can be a reference path (e.g. "$.platform_capabilities")
            job_role_arn (str | None): Job role arn to use for the job. This can be a reference path (e.g. "$.job_role_arn")
        """  # noqa: E501
        super().__init__(scope, id, env_base)
        key_prefix = key_prefix or S3_SCRATCH_KEY_PREFIX

        request_key = sfn.JsonPath.format(
            f"{key_prefix}{{}}/{{}}/request.json",
            sfn.JsonPath.execution_name,
            sfn.JsonPath.string_at("$.taskResult.prep.task_id"),
        )
        response_key = sfn.JsonPath.format(
            f"{key_prefix}{{}}/{{}}/response.json",
            sfn.JsonPath.execution_name,
            sfn.JsonPath.string_at("$.taskResult.prep.task_id"),
        )

        start = sfn.Pass(
            self,
            f"{id} Prep S3 Keys",
            parameters={
                "task_id": sfn.JsonPath.uuid(),
            },
            result_path="$.taskResult.prep",
        )

        if mount_point_configs:
            if mount_points or volumes:
                raise ValueError("Cannot specify both mount_point_configs and mount_points")
            mount_points, volumes = self.convert_to_mount_point_and_volumes(mount_point_configs)

        put_payload = S3Operation.put_payload(
            self,
            f"{id} Put Request to S3",
            payload=payload_path or sfn.JsonPath.entire_payload,
            bucket_name=bucket_name,
            key=request_key,
            result_path="$.taskResult.put",
        )

        submit_job = SubmitJobFragment(
            self,
            id + "Batch",
            env_base=env_base,
            name=name,
            job_queue=job_queue,
            command=[
                "run_cli_executor",
                "--executor",
                executor,
                "--input",
                sfn.JsonPath.format("s3://{}/{}", "$.Bucket", "$.Key"),
                "--output-location",
                sfn.JsonPath.format("s3://{}/{}", bucket_name, response_key),
            ],
            image=image,
            environment=environment,
            memory=memory,
            vcpus=vcpus,
            mount_points=mount_points or [],
            volumes=volumes or [],
            platform_capabilities=platform_capabilities,
            job_role_arn=job_role_arn,
        )

        get_response = S3Operation.get_payload(
            self,
            f"{id}",
            bucket_name=bucket_name,
            key=response_key,
        ).to_single_state(
            f"{id} Get Response from S3",
            output_path="$[0]",
        )

        self.definition = start.next(put_payload).next(submit_job).next(get_response)

    @property
    def start_state(self) -> sfn.State:
        return self.definition.start_state

    @property
    def end_states(self) -> list[sfn.INextable]:
        return self.definition.end_states



"""Base construct classes and mixins for environment-aware CDK constructs.

This module provides the foundational classes and mixins that enable
environment-aware CDK constructs throughout the library.
"""

import hashlib
import re

import aws_cdk as cdk
from aibs_informatics_core.env import EnvBase, EnvBaseMixins, EnvType, ResourceNameBaseEnum
from aws_cdk import Stack
from aws_cdk import aws_iam as iam
from constructs import Construct

from aibs_informatics_cdk_lib.common.aws.iam_utils import grant_managed_policies


class EnvBaseConstructMixins(EnvBaseMixins):
    """Mixin class providing environment-aware utilities for CDK constructs.

    This mixin extends EnvBaseMixins with CDK-specific functionality including
    resource naming, tagging, and IAM utilities.
    """

    @property
    def is_dev(self) -> bool:
        return self.env_base.env_type == EnvType.DEV

    @property
    def is_test(self) -> bool:
        return self.env_base.env_type == EnvType.TEST

    @property
    def is_prod(self) -> bool:
        return self.env_base.env_type == EnvType.PROD

    @property
    def is_test_or_prod(self) -> bool:
        return self.is_prod or self.env_base.env_type == EnvType.TEST

    @property
    def construct_tags(self) -> list[cdk.Tag]:
        return []

    @property
    def aws_region(self) -> str:
        return cdk.Stack.of(self.as_construct()).region

    @property
    def aws_account(self) -> str:
        return cdk.Stack.of(self.as_construct()).account

    def add_tags(self, *tags: cdk.Tag):
        """Add tags to the construct.

        Args:
            *tags (cdk.Tag): Tags to add to the construct.
        """
        for tag in tags:
            cdk.Tags.of(self.as_construct()).add(key=tag.key, value=tag.value)

    def normalize_construct_id(
        self, construct_id: str, max_size: int = 64, hash_size: int = 8
    ) -> str:
        """Normalize a construct ID to fit within size constraints.

        Replaces special characters with dashes and truncates long IDs
        by appending a hash suffix.

        Args:
            construct_id (str): The original construct ID.
            max_size (int): Maximum allowed length. Defaults to 64.
            hash_size (int): Length of hash suffix for truncation. Defaults to 8.

        Returns:
            The normalized construct ID.

        Raises:
            ValueError: If max_size is less than hash_size.
        """
        if max_size < hash_size:
            raise ValueError("max_size must be greater than hash_size: ")

        # Replace special characters with dashes
        string = re.sub(r"\W+", "-", construct_id)

        # Check if the string exceeds the allowed size
        if len(string) > max_size:
            # Generate a hexdigest of the string

            digest = hashlib.sha256(string.encode()).hexdigest()
            digest = digest[:hash_size]
            string = string[: -len(digest)] + digest

        return string

    def get_construct_id(self, *names: str) -> str:
        """Build a construct ID from name components.

        Args:
            *names (str): Name components to include in the construct ID.

        Returns:
            The constructed ID with environment prefix.
        """
        return self.build_construct_id(self.env_base, *names)

    def get_child_id(self, scope: Construct, *names: str) -> str:
        """Build a child construct ID, avoiding duplicate env_base prefixes.

        Args:
            scope (Construct): The parent construct scope.
            *names (str): Name components for the child ID.

        Returns:
            The child construct ID.
        """
        if scope.node.id.startswith(self.env_base):
            return "-".join(names)
        return self.build_construct_id(self.env_base, *names)

    def get_name_with_env(self, *names: str) -> str:
        """Get a name prefixed with the environment base.

        Args:
            *names (str): Name components to prefix.

        Returns:
            The environment-prefixed name.
        """
        return self.env_base.prefixed(*names)

    def get_resource_name(self, name: ResourceNameBaseEnum | str) -> str:
        """Get the full resource name with environment prefix.

        Args:
            name (Union[ResourceNameBaseEnum, str]): The resource name or enum.

        Returns:
            The fully qualified resource name.
        """
        if isinstance(name, ResourceNameBaseEnum):
            return name.get_name(self.env_base)
        return self.env_base.get_resource_name(name)

    def get_stack_of(self, construct: Construct | None = None) -> Stack:
        """Get the stack containing a construct.

        Args:
            construct (Optional[Construct]): The construct to find the stack for.
                Defaults to self.

        Returns:
            The CDK Stack containing the construct.
        """
        if construct is None:
            construct = self.as_construct()
        return cdk.Stack.of(construct)

    def as_construct(self) -> Construct:
        """Return this object as a Construct.

        Returns:
            This object cast as a Construct.

        Raises:
            AssertionError: If this object is not a Construct instance.
        """
        assert isinstance(self, Construct)
        return self

    @classmethod
    def build_construct_id(cls, env_base: EnvBase, *names: str) -> str:
        """Build a construct ID from environment base and name components.

        Args:
            env_base (EnvBase): The environment base configuration.
            *names (str): Name components to include in the ID.

        Returns:
            The constructed ID string.
        """
        return env_base.get_construct_id(*names)  # , cls.__name__)

    # ---------------------------------------------------------------
    #                       Resource Helpers
    # ---------------------------------------------------------------

    # ----------
    #   IAM

    @classmethod
    def add_managed_policies(
        cls, role: iam.IRole | None, *managed_policies: str | iam.ManagedPolicy
    ):
        """Add managed policies to an IAM role.

        Args:
            role (Optional[iam.IRole]): The IAM role to add policies to.
            *managed_policies (Union[str, iam.ManagedPolicy]): Managed policies
                to add, either by name or as ManagedPolicy objects.
        """
        grant_managed_policies(role, *managed_policies)


class EnvBaseConstruct(Construct, EnvBaseConstructMixins):
    """Base construct class with environment awareness.

    Provides a foundation for creating CDK constructs that are aware of
    the deployment environment and support automatic tagging.
    """

    def __init__(self, scope: Construct, id: str | None, env_base: EnvBase) -> None:
        """Initialize an environment-aware construct.

        Args:
            scope (Construct): The parent construct scope.
            id (Optional[str]): The construct ID. Defaults to class name if None.
            env_base (EnvBase): The environment base configuration.
        """
        super().__init__(scope, id or self.__class__.__name__)
        self.env_base = env_base
        self.add_tags(*self.construct_tags)

    @property
    def construct_id(self) -> str:
        return self.node.id



__all__ = [
    "EFSEcosystem",
    "EnvBaseFileSystem",
    "grant_connectable_file_system_access",
    "grant_role_file_system_access",
    "grant_file_system_access",
]

from .file_system import (
    EFSEcosystem,
    EnvBaseFileSystem,
    grant_connectable_file_system_access,
    grant_file_system_access,
    grant_role_file_system_access,
)



from collections.abc import Mapping, Sequence
from typing import Any

import aws_cdk as cdk
import constructs
from aibs_informatics_core.env import EnvBase, EnvType, ResourceNameBaseEnum
from aws_cdk import aws_cloudwatch as cloudwatch
from aws_cdk import aws_lambda as lambda_
from aws_cdk import aws_sns as sns
from aws_cdk import aws_sqs as sqs
from aws_cdk.aws_lambda_event_sources import SqsEventSource
from aws_cdk.aws_sns_subscriptions import SqsSubscription


class ExternalSnsTrigger(constructs.Construct):
    """This intended to be a generic CDK construct that defines resources necessary to
    implement a trigger system that listens for events from an external
    (some other AWS account) SNS Topic and fires off a lambda in the
    account where this stack is deployed.

    ``` Diagram:
                                ┌---- ExternalSNSTrigger ----┐
                                |                            |
        (other AWS account)     |           (SQS)            | (optional provided Lambda)
        external_sns_topic -----> external_sns_event_queue -----> triggered_lambda_fn
                                |             |              |
                                |             v              |
                                |         (SQS DLQ)          |
                                |   external_sns_event_dlq   |
                                |             |              |
                                |             v              |
                                |        (Cloudwatch)        |
                                | triggered_lambda_dlq_alarm |
                                |                            |
                                └----------------------------┘

    ```

    The main intended use case is to provide a simple template for setting up automation
    based on PTS SNS Topic notifications in a separate AWS account (e.g. in an informatics processing pipeline)

    Example usage:
    self.external_sns_trigger_construct = ExternalSnsTrigger(
        scope=self,
        id=self.env_base.get_construct_id("merscope-imaging-process-sns-trigger"),
        env_base=self.env_base,
        triggered_lambda_fn=self.pts_listener_fn,
        external_sns_event_name="merscope-imaging-process",
        external_sns_topic_arn=f"arn:aws:sns:us-west-2:{account_id}:{sns_topic_name}",
    )

    NOTE: The `triggered_lambda_fn` is optional and if you have an alternative arrangement
          for triggering off of SQS messages (e.g. Airflow SQS Sensor) you can provide None as the
          `triggered_lambda_fn` when instantiating the ExternalSnsTrigger construct.
    """  # noqa: E501

    @property
    def queue_name(self) -> str:
        if self._external_sns_event_queue_name is not None:
            return self._env_base.prefixed(self._external_sns_event_queue_name)
        return self._env_base.prefixed(self._external_sns_event_name, "sns-event-queue")

    @property
    def dlq_name(self) -> str:
        if self._external_sns_event_dlq_name is not None:
            return self._env_base.prefixed(self._external_sns_event_dlq_name)
        return self._env_base.prefixed(self._external_sns_event_name, "sns-event-dlq")

    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        triggered_lambda_fn: lambda_.Function | None,
        external_sns_event_name: str | ResourceNameBaseEnum,
        external_sns_topic_arn: str,
        external_sns_event_queue_filters: Sequence[Mapping[str, Any]] | None = None,
        external_sns_event_queue_name: str | None = None,
        external_sns_event_dlq_name: str | None = None,
        external_sns_event_queue_retention_period: cdk.Duration | None = cdk.Duration.days(7),
        sqs_event_source_enabled: bool | None = None,
        **kwargs,
    ) -> None:
        super().__init__(scope=scope, id=id)

        self._env_base = env_base
        self._external_sns_event_name = external_sns_event_name
        self._external_sns_event_queue_name = external_sns_event_queue_name
        self._external_sns_event_dlq_name = external_sns_event_dlq_name

        if sqs_event_source_enabled is None:
            sqs_event_source_enabled = env_base.env_type is EnvType.PROD

        self.external_sns_event_dlq = sqs.Queue(
            scope=self,
            id=env_base.get_construct_id(external_sns_event_name, "sns-event-dlq"),
            queue_name=self.dlq_name,
            retention_period=cdk.Duration.days(14),
        )

        self.external_sns_event_queue = sqs.Queue(
            scope=self,
            id=env_base.get_construct_id(external_sns_event_name, "sns-event-queue"),
            queue_name=self.queue_name,
            retention_period=external_sns_event_queue_retention_period,
            dead_letter_queue=sqs.DeadLetterQueue(
                max_receive_count=2,
                queue=self.external_sns_event_dlq,
            ),
            # visibility_timeout must be longer than the `timeout` of the
            # `triggered_lambda_fn` or deploy will fail
            visibility_timeout=cdk.Duration.seconds(330),
        )

        # The *owning* AWS account of the SNS topic needs to give Subscribe permissions to the
        # AWS account where this stack will be deployed
        # See: https://docs.aws.amazon.com/sns/latest/dg/sns-send-message-to-sqs-cross-account.html
        self.external_sns_topic = sns.Topic.from_topic_arn(
            scope=self,
            id=env_base.get_construct_id(external_sns_event_name, "external-sns-topic"),
            topic_arn=external_sns_topic_arn,
        )

        # Useful reference:
        # https://docs.aws.amazon.com/cdk/api/v2/python/aws_cdk.aws_sns/Subscription.html#aws_cdk.aws_sns.Subscription
        self.external_sns_topic.add_subscription(
            SqsSubscription(
                queue=self.external_sns_event_queue,
                raw_message_delivery=True,
            )
        )

        if triggered_lambda_fn is not None:
            triggered_lambda_fn.add_event_source(
                source=SqsEventSource(
                    queue=self.external_sns_event_queue,
                    report_batch_item_failures=True,
                    filters=external_sns_event_queue_filters,
                    enabled=sqs_event_source_enabled,
                )
            )
            self.external_sns_event_queue.grant_consume_messages(triggered_lambda_fn)

        # Alarm that fires if external_sns_event_queue fails delivery or if lambda fails to process
        # Further actions can be configured by accessing ExternalSnsTrigger.triggered_lambda_dlq_alarm resource # noqa: E501
        # TODO: Revisit making alarms even more configurable in the future
        self.triggered_lambda_dlq_alarm = cloudwatch.Alarm(
            scope=self,
            id=env_base.get_construct_id(external_sns_event_name, "sns-event-dlq-alarm"),
            alarm_description=(
                f"Alarm if more than 1 message in {self.dlq_name} in 10 minute period"
            ),
            metric=self.external_sns_event_dlq.metric_approximate_number_of_messages_visible(
                statistic=cloudwatch.Stats.MAXIMUM,
                period=cdk.Duration.minutes(10),
            ),
            evaluation_periods=1,
            threshold=1,
            datapoints_to_alarm=1,
            comparison_operator=cloudwatch.ComparisonOperator.GREATER_THAN_OR_EQUAL_TO_THRESHOLD,
        )



from typing import Literal, TypedDict

from aws_cdk import aws_cloudwatch as cw


def to_comparison_operator(value: cw.ComparisonOperator | str) -> cw.ComparisonOperator:
    if isinstance(value, cw.ComparisonOperator):
        return value
    elif value.lower() in [">=", "greater_than_or_equal_to"]:
        return cw.ComparisonOperator.GREATER_THAN_OR_EQUAL_TO_THRESHOLD
    elif value.lower() in [">", "greater_than"]:
        return cw.ComparisonOperator.GREATER_THAN_THRESHOLD
    elif value.lower() in ["<=", "less_than_or_equal_to"]:
        return cw.ComparisonOperator.LESS_THAN_OR_EQUAL_TO_THRESHOLD
    elif value.lower() in ["<", "less_than"]:
        return cw.ComparisonOperator.LESS_THAN_THRESHOLD
    elif value.lower() in ["<>", "out_of_range"]:
        return cw.ComparisonOperator.LESS_THAN_LOWER_OR_GREATER_THAN_UPPER_THRESHOLD
    else:
        return cw.ComparisonOperator(value)


class _AlarmMetricConfigOptional(TypedDict, total=False):
    pass


class _AlarmMetricConfigRequired(TypedDict):
    name: str
    threshold: int
    evaluation_periods: int
    datapoints_to_alarm: int
    comparison_operator: cw.ComparisonOperator | str


class AlarmMetricConfig(_AlarmMetricConfigRequired, _AlarmMetricConfigOptional):
    pass


class _GraphMetricConfigOptional(TypedDict, total=False):
    namespace: str
    metric_expression: str
    dimension_map: dict[str, str]
    using_metrics: dict[str, cw.IMetric]
    alarm: AlarmMetricConfig
    axis_side: Literal["left", "right"]
    color: str
    label: str
    unit: cw.Unit


class _GraphMetricConfigRequired(TypedDict):
    metric: str | cw.IMetric
    statistic: str


class GraphMetricConfig(_GraphMetricConfigRequired, _GraphMetricConfigOptional):
    pass


class _GroupedGraphMetricConfigOptional(TypedDict, total=False):
    namespace: str
    dimension_map: dict[str, str]
    height: int
    width: int
    left_y_axis: cw.YAxisProps
    right_y_axis: cw.YAxisProps


class _GroupedGraphMetricConfigRequired(TypedDict):
    title: str
    metrics: list[GraphMetricConfig]


class GroupedGraphMetricConfig(
    _GroupedGraphMetricConfigRequired, _GroupedGraphMetricConfigOptional
):
    pass



import re
from collections import defaultdict
from copy import deepcopy
from math import ceil
from typing import Any, Literal

import aws_cdk as cdk
import constructs
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_cloudwatch as cw
from aws_cdk import aws_cloudwatch_actions as cw_actions
from aws_cdk import aws_sns as sns

from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins
from aibs_informatics_cdk_lib.constructs_.cw.types import (
    GroupedGraphMetricConfig,
    to_comparison_operator,
)


class DashboardMixins(EnvBaseConstructMixins):
    @property
    def dashboard(self) -> cw.Dashboard:
        return self._dashboard

    @dashboard.setter
    def dashboard(self, dashboard: cw.Dashboard) -> None:
        self._dashboard = dashboard

    def add_graphs(
        self,
        grouped_metric_configs: list[GroupedGraphMetricConfig],
        namespace: str,
        period: cdk.Duration,
        alarm_id_discriminator: str,
        alarm_topic: sns.Topic | None,
        dimensions: dict[str, Any],
    ) -> None:
        """Adds graphs to a dashboard based on metrics configs.

        Args:
            grouped_metric_configs (List[GroupedGraphMetricConfig]): config
            namespace (str): default namespace for metrics, can be overridden in configs
            period (cdk.Duration): default duration
            alarm_id_discriminator (str): a non token disciminator for alarms
            alarm_name_discriminator (str): common name alarm discriminator
            alarm_topic (Optional[sns.Topic]): optional SNS topic to send alarm notifications
            dimensions (Dict[str, Any]): dimensions to generate metrics from
        """
        # First, calculate widths dynamically
        MAX_PER_ROW = 4
        TOTAL_WIDTH = 24
        grouped_metric_configs = deepcopy(grouped_metric_configs)
        for idx in range(0, len(grouped_metric_configs), MAX_PER_ROW):
            grouped_metric_configs_subset = grouped_metric_configs[idx : idx + MAX_PER_ROW]
            requested_widget_widths = [_.get("width", 0) for _ in grouped_metric_configs_subset]
            remaining_width = TOTAL_WIDTH - sum(requested_widget_widths)
            widgets_without_width = sum([_ == 0 for _ in requested_widget_widths])
            if not remaining_width or not widgets_without_width:
                continue
            default_widget_width = remaining_width // widgets_without_width
            for grouped_metric_config in grouped_metric_configs_subset:
                grouped_metric_config["width"] = default_widget_width

        # Next generate the graph widgets and alarms
        graph_widgets, metric_alarms = self.create_widgets_and_alarms(
            grouped_metric_configs=grouped_metric_configs,
            namespace=namespace,
            period=period,
            alarm_id_discriminator=alarm_id_discriminator,
            alarm_topic=alarm_topic,
            dimensions=dimensions,
        )

        for idx in range(0, len(graph_widgets), MAX_PER_ROW):
            self.dashboard.add_widgets(*graph_widgets[idx : idx + MAX_PER_ROW])
        if metric_alarms:
            max_alarms_per_row = 6  # This is how many fit with full screen (improve me)
            num_alarms = len(metric_alarms)
            alarm_widget_height = ceil(ceil(num_alarms // max_alarms_per_row) * 1.5)
            self.dashboard.add_widgets(
                cw.AlarmStatusWidget(
                    alarms=metric_alarms,
                    height=alarm_widget_height,
                    width=24,
                )
            )

    def create_widgets_and_alarms(
        self,
        grouped_metric_configs: list[GroupedGraphMetricConfig],
        namespace: str,
        period: cdk.Duration,
        alarm_id_discriminator: str,
        alarm_topic: sns.Topic | None,
        dimensions: dict[str, Any],
    ) -> tuple[list[cw.IWidget], list[cw.IAlarm]]:
        """Create graph widgets and alarms from configs

        Args:
            grouped_metric_configs (List[GroupedGraphMetricConfig]): configs
            namespace (str): default metric namespace
            period (cdk.Duration): default duration
            alarm_name_discriminator (str): alarm discriminator name
            alarm_topic (Optional[sns.Topic]): optional sns topic for alarms
            dimensions (Dict[str, Any]):

        Returns:
            Tuple[List[cw.IWidget], List[cw.IAlarm]]: List of widgets and list of alarms
        """
        self_stack = cdk.Stack.of(self.dashboard)

        graph_widgets: list[cw.IWidget] = []
        metric_alarms: list[cw.IAlarm] = []
        for grouped_metric_config in grouped_metric_configs:
            lr_graph_metrics: dict[Literal["left", "right"], list[cw.Metric]] = defaultdict(list)
            lr_annotations: dict[Literal["left", "right"], list[cw.HorizontalAnnotation]] = (
                defaultdict(list)
            )

            graph_metric_namespace = grouped_metric_config.get("namespace", namespace)
            graph_dimension_map = {**dimensions, **grouped_metric_config.get("dimension_map", {})}
            for metric_config in grouped_metric_config["metrics"]:
                if isinstance(metric_config["metric"], (cw.Metric, cw.MathExpression)):
                    graph_metric = metric_config["metric"]
                    metric_label = metric_config.get("label", graph_metric.label)
                    if isinstance(graph_metric, cw.Metric):
                        metric_name = graph_metric.metric_name
                    else:
                        metric_name = graph_metric.label or metric_config["statistic"]
                else:
                    metric = metric_config["metric"]
                    if isinstance(metric, cw.Metric):
                        metric_name = metric.metric_name
                    else:
                        metric_name = str(metric)
                    metric_label = metric_config.get(
                        "label",
                        re.sub(
                            r"([a-z])([A-Z])",
                            r"\1 \2",
                            re.sub(r"([A-Z])([a-z])", r" \1\2", metric_name.replace(".", " ")),
                        ),
                    )

                    metric_expression = metric_config.get("metric_expression")
                    if metric_expression:
                        graph_metric = cw.MathExpression(
                            expression=metric_expression,
                            using_metrics=metric_config.get("using_metrics", {}),
                            label=metric_label,
                        )
                    else:
                        graph_metric = cw.Metric(
                            metric_name=metric_name,
                            namespace=metric_config.get("namespace", graph_metric_namespace),
                            label=metric_label,
                            statistic=metric_config["statistic"],
                            period=period,
                            dimensions_map={
                                **graph_dimension_map,
                                **metric_config.get("dimension_map", {}),
                            },
                            unit=metric_config.get("unit"),
                        )

                metric_axis = metric_config.get("axis_side", "left")
                lr_graph_metrics[metric_axis].append(graph_metric)  # type: ignore # MathExpression implements IMetric

                metric_alarm_config = metric_config.get("alarm")
                if metric_alarm_config:
                    alarm_name = metric_alarm_config["name"]
                    alarm = graph_metric.create_alarm(
                        self_stack,
                        self.get_construct_id(alarm_name, alarm_id_discriminator),
                        # TODO: every time a change is made to these alarms, Cfn throws an error
                        #       for trying to modify what is a custom resource. So instead, let
                        #       the name be autogenerated.
                        # alarm_name=alarm_name,
                        alarm_description=f"Alarm for {alarm_name}",
                        threshold=metric_alarm_config["threshold"],
                        evaluation_periods=metric_alarm_config["evaluation_periods"],
                        datapoints_to_alarm=metric_alarm_config["datapoints_to_alarm"],
                        comparison_operator=to_comparison_operator(
                            metric_alarm_config["comparison_operator"]
                        ),
                    )
                    lr_annotations[metric_axis].append(
                        cw.HorizontalAnnotation(
                            value=metric_alarm_config["threshold"],
                            color=graph_metric.color,
                        )
                    )
                    metric_alarms.append(alarm)
                    if alarm_topic:
                        alarm.add_alarm_action(cw_actions.SnsAction(alarm_topic))  # type: ignore # SnsAction implements IAlarmAction

            graph_widgets.append(
                cw.GraphWidget(
                    title=grouped_metric_config["title"],
                    left=lr_graph_metrics["left"],
                    left_annotations=lr_annotations["left"],
                    left_y_axis=grouped_metric_config.get("left_y_axis"),
                    right=lr_graph_metrics["right"],
                    right_annotations=lr_annotations["right"],
                    right_y_axis=grouped_metric_config.get("right_y_axis"),
                    height=grouped_metric_config.get("height", 10),
                    width=grouped_metric_config.get("width"),
                )
            )
        return graph_widgets, metric_alarms

    def add_text_widget(
        self,
        header: str,
        header_level: int | None,
        body: str | None = None,
        height: int = 2,
        width: int = 24,
    ):
        self.dashboard.add_widgets(
            self.build_text_widget(
                header=header, header_level=header_level, body=body, height=height, width=width
            )
        )

    @classmethod
    def build_text_widget(
        cls,
        header: str,
        header_level: int | None = None,
        body: str | None = None,
        height: int = 2,
        width: int = 24,
    ) -> cw.TextWidget:
        markdown = f"{'#' * max(min(header_level, 6), 1) if header_level else ''} {header}"
        if body:
            markdown += f"\n{body}\n"
        return cw.TextWidget(markdown=markdown, height=height, width=width)


class EnhancedDashboard(DashboardMixins, cw.Dashboard):
    def __init__(
        self,
        scope: constructs.Construct,
        id: str | None,
        env_base: EnvBase,
        dashboard_name: str,
        **dashboard_kwargs,
    ) -> None:
        self.env_base = env_base
        super().__init__(
            scope,
            id or self.get_name_with_env(dashboard_name),
            dashboard_name=self.get_name_with_env(dashboard_name),
            **dashboard_kwargs,
        )
        self.dashboard = self


class DashboardTools(DashboardMixins):
    def __init__(self, dashboard: cw.Dashboard) -> None:
        self.dashboard = dashboard



"""Step Functions state machine fragment constructs.

This module provides base classes and utilities for building Step Functions
state machine fragments and workflows.
"""

from abc import abstractmethod
from collections.abc import Mapping, Sequence
from typing import Any, TypeVar, cast

import aws_cdk as cdk
import constructs
from aibs_informatics_core.collections import ValidatedStr
from aibs_informatics_core.env import EnvBase
from aws_cdk import aws_iam as iam
from aws_cdk import aws_lambda as lambda_
from aws_cdk import aws_logs as logs_
from aws_cdk import aws_stepfunctions as sfn

from aibs_informatics_cdk_lib.common.aws.core_utils import build_lambda_arn
from aibs_informatics_cdk_lib.common.aws.sfn_utils import JsonReferencePath
from aibs_informatics_cdk_lib.constructs_.base import EnvBaseConstructMixins

T = TypeVar("T", bound=ValidatedStr)

F = TypeVar("F", bound="StateMachineFragment")


def create_log_options(
    scope: constructs.Construct,
    id: str,
    env_base: EnvBase,
    removal_policy: cdk.RemovalPolicy | None = None,
    retention: logs_.RetentionDays | None = None,
) -> sfn.LogOptions:
    """Create log options for a state machine.

    Args:
        scope (constructs.Construct): The construct scope.
        id (str): Identifier for the log group.
        env_base (EnvBase): Environment base for naming.
        removal_policy (Optional[cdk.RemovalPolicy]): Removal policy.
            Defaults to DESTROY.
        retention (Optional[logs_.RetentionDays]): Log retention period.
            Defaults to ONE_MONTH.

    Returns:
        Log options configured with a CloudWatch log group.
    """
    return sfn.LogOptions(
        destination=logs_.LogGroup(
            scope,
            env_base.get_construct_id(id, "state-loggroup"),
            log_group_name=env_base.get_state_machine_log_group_name(id),
            removal_policy=removal_policy or cdk.RemovalPolicy.DESTROY,
            retention=retention or logs_.RetentionDays.ONE_MONTH,
        )
    )


def create_role(
    scope: constructs.Construct,
    id: str,
    env_base: EnvBase,
    assumed_by: iam.IPrincipal = iam.ServicePrincipal("states.amazonaws.com"),  # type: ignore[assignment]
    managed_policies: Sequence[iam.IManagedPolicy | str] | None = None,
    inline_policies: Mapping[str, iam.PolicyDocument] | None = None,
    inline_policies_from_statements: Mapping[str, Sequence[iam.PolicyStatement]] | None = None,
    include_default_managed_policies: bool = True,
) -> iam.Role:
    """Create an IAM role for a state machine.

    Args:
        scope (constructs.Construct): The construct scope.
        id (str): Identifier for the role.
        env_base (EnvBase): Environment base for naming.
        assumed_by (iam.IPrincipal): Principal that can assume the role.
            Defaults to states.amazonaws.com.
        managed_policies (Optional[Sequence[Union[iam.IManagedPolicy, str]]]):
            Managed policies to attach.
        inline_policies (Optional[Mapping[str, iam.PolicyDocument]]):
            Inline policy documents.
        inline_policies_from_statements (Optional[Mapping[str, Sequence[iam.PolicyStatement]]]):
            Inline policies from statements.
        include_default_managed_policies (bool): Include default Step Functions
            managed policies. Defaults to True.

    Returns:
        The created IAM role.
    """
    construct_id = env_base.get_construct_id(id, "role")

    if managed_policies is not None:
        managed_policies = [
            iam.ManagedPolicy.from_aws_managed_policy_name(policy)
            if isinstance(policy, str)
            else policy
            for policy in managed_policies
        ]

    if inline_policies is None:
        inline_policies = {}
    if inline_policies_from_statements:
        inline_policies = {
            **inline_policies,
            **{
                name: iam.PolicyDocument(statements=statements)
                for name, statements in inline_policies_from_statements.items()
            },
        }

    return iam.Role(
        scope,
        construct_id,
        assumed_by=assumed_by,  # type: ignore
        managed_policies=[
            *(managed_policies or []),  # type: ignore[list-item] # mypy does not understand that ManagedPolicy is an IManagedPolicy
            *[
                iam.ManagedPolicy.from_aws_managed_policy_name(policy)
                for policy in (
                    [
                        "AWSStepFunctionsFullAccess",
                        "CloudWatchLogsFullAccess",
                        "CloudWatchEventsFullAccess",
                    ]
                    if include_default_managed_policies
                    else []
                )
            ],
        ],
        inline_policies=inline_policies,
    )


class StateMachineMixins(EnvBaseConstructMixins):
    """Mixin class providing state machine helper methods.

    Provides methods for retrieving Lambda functions and state machines
    with caching support.
    """

    def get_fn(self, function_name: str) -> lambda_.IFunction:
        """Get a Lambda function by name.

        Args:
            function_name (str): The function name.

        Returns:
            The Lambda function interface.
        """
        cache_attr = "_function_cache"
        if not hasattr(self, cache_attr):
            setattr(self, cache_attr, {})
        resource_cache = cast(dict[str, lambda_.IFunction], getattr(self, cache_attr))
        if function_name not in resource_cache:
            resource_cache[function_name] = lambda_.Function.from_function_arn(
                scope=self.as_construct(),
                id=self.env_base.get_construct_id(function_name, "from-arn"),
                function_arn=build_lambda_arn(
                    resource_type="function",
                    resource_id=self.env_base.get_function_name(function_name),
                ),
            )
        return resource_cache[function_name]

    def get_state_machine_from_name(self, state_machine_name: str) -> sfn.IStateMachine:
        """Get a state machine by name.

        Args:
            state_machine_name (str): The state machine name.

        Returns:
            The state machine interface.
        """
        cache_attr = "_state_machine_cache"
        if not hasattr(self, cache_attr):
            setattr(self, cache_attr, {})
        resource_cache = cast(dict[str, sfn.IStateMachine], getattr(self, cache_attr))
        if state_machine_name not in resource_cache:
            resource_cache[state_machine_name] = sfn.StateMachine.from_state_machine_name(
                scope=self.as_construct(),
                id=self.env_base.get_construct_id(state_machine_name, "from-name"),
                state_machine_name=self.env_base.get_state_machine_name(state_machine_name),
            )
        return resource_cache[state_machine_name]


def create_state_machine(
    scope: constructs.Construct,
    env_base: EnvBase,
    id: str,
    name: str | None,
    definition: sfn.IChainable,
    role: iam.Role | None = None,
    logs: sfn.LogOptions | None = None,
    timeout: cdk.Duration | None = None,
) -> sfn.StateMachine:
    """Create a state machine from a definition.

    Args:
        scope (constructs.Construct): The construct scope.
        env_base (EnvBase): Environment base for naming.
        id (str): Construct identifier.
        name (Optional[str]): State machine name.
        definition (sfn.IChainable): The state machine definition.
        role (Optional[iam.Role]): IAM role for the state machine.
        logs (Optional[sfn.LogOptions]): Log options.
        timeout (Optional[cdk.Duration]): Execution timeout.

    Returns:
        The created state machine.
    """
    return sfn.StateMachine(
        scope,
        env_base.get_construct_id(id),
        state_machine_name=env_base.get_state_machine_name(name) if name else None,
        logs=(
            logs
            or sfn.LogOptions(
                destination=logs_.LogGroup(
                    scope,
                    env_base.get_construct_id(id, "state-loggroup"),
                    log_group_name=env_base.get_state_machine_log_group_name(name or id),
                    removal_policy=cdk.RemovalPolicy.DESTROY,
                    retention=logs_.RetentionDays.ONE_MONTH,
                )
            )
        ),
        role=cast(iam.IRole, role),
        definition_body=sfn.DefinitionBody.from_chainable(definition),
        timeout=timeout,
    )


class StateMachineFragment(sfn.StateMachineFragment):
    """Base class for state machine fragments.

    Provides common functionality for building reusable state machine
    fragments with definition management.
    """

    @property
    def definition(self) -> sfn.IChainable:
        """Get the state machine definition.

        Returns:
            The chainable definition.
        """
        return self._definition

    @definition.setter
    def definition(self, value: sfn.IChainable):
        """Set the state machine definition.

        Args:
            value (sfn.IChainable): The definition to set.
        """
        self._definition = value

    @property
    def start_state(self) -> sfn.State:
        """Get the start state.

        Returns:
            The definition's start state.
        """
        return self.definition.start_state

    @property
    def end_states(self) -> list[sfn.INextable]:
        """Get the end states.

        Returns:
            List of nextable end states.
        """
        return self.definition.end_states

    def enclose(
        self,
        id: str | None = None,
        input_path: str | None = None,
        result_path: str | None = None,
    ) -> sfn.Chain:
        """Enclose the fragment within a parallel state.

        Args:
            id (Optional[str]): Identifier for the parallel state.
                Defaults to the node ID.
            input_path (Optional[str]): Input path for the enclosed state.
                Defaults to "$".
            result_path (Optional[str]): Result path for output.
                Defaults to same as input_path.

        Returns:
            The enclosed state machine fragment as a chain.
        """
        id = id or self.node.id

        if input_path is None:
            input_path = "$"
        if result_path is None:
            result_path = input_path

        chain = (
            sfn.Chain.start(self.definition)
            if not isinstance(self.definition, (sfn.Chain, sfn.StateMachineFragment))
            else self.definition
        )

        if isinstance(chain, sfn.Chain):
            parallel = chain.to_single_state(
                id=f"{id} Enclosure", input_path=input_path, result_path=result_path
            )
        else:
            parallel = chain.to_single_state(input_path=input_path, result_path=result_path)
        definition = sfn.Chain.start(parallel)

        if result_path and result_path != sfn.JsonPath.DISCARD:
            restructure = sfn.Pass(
                self,
                f"{id} Enclosure Post",
                input_path=f"{result_path}[0]",
                result_path=result_path,
            )
            definition = definition.next(restructure)

        return definition


class EnvBaseStateMachineFragment(StateMachineFragment, StateMachineMixins):
    """Environment-aware state machine fragment.

    Combines StateMachineFragment with environment base naming conventions
    and state machine mixins.
    """

    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
    ) -> None:
        """Initialize an environment-aware state machine fragment.

        Args:
            scope (constructs.Construct): The construct scope.
            id (str): The construct ID.
            env_base (EnvBase): Environment base for resource naming.
        """
        super().__init__(scope, id)
        self.env_base = env_base

    def to_single_state(
        self,
        *,
        prefix_states: str | None = None,
        state_id: str | None = None,
        comment: str | None = None,
        input_path: str | None = None,
        output_path: str | None = "$[0]",
        result_path: str | None = None,
        result_selector: Mapping[str, Any] | None = None,
        arguments: Mapping[str, Any] | None = None,
        parameters: Mapping[str, Any] | None = None,
        query_language: sfn.QueryLanguage | None = None,
        state_name: str | None = None,
        assign: Mapping[str, Any] | None = None,
        outputs: Any = None,
    ) -> sfn.Parallel:
        """Convert the fragment to a single parallel state.

        Args:
            prefix_states (Optional[str]): Prefix for state names.
            state_id (Optional[str]): ID for the parallel state.
            comment (Optional[str]): Comment for the state.
            input_path (Optional[str]): Input path.
            output_path (Optional[str]): Output path. Defaults to "$[0]".
            result_path (Optional[str]): Result path.
            result_selector (Optional[Mapping[str, Any]]): Result selector.

        Returns:
            A parallel state containing the fragment.
        """
        return super().to_single_state(
            prefix_states=prefix_states,
            state_id=state_id,
            comment=comment,
            input_path=input_path,
            output_path=output_path,
            result_path=result_path,
            result_selector=result_selector,
            arguments=arguments,
            parameters=parameters,
            query_language=query_language,
            state_name=state_name,
            assign=assign,
            outputs=outputs,
        )

    def to_state_machine(
        self,
        state_machine_name: str,
        role: iam.Role | None = None,
        logs: sfn.LogOptions | None = None,
        timeout: cdk.Duration | None = None,
    ) -> sfn.StateMachine:
        """Convert the fragment to a complete state machine.

        Args:
            state_machine_name (str): Name for the state machine.
            role (Optional[iam.Role]): IAM role. Creates default if None.
            logs (Optional[sfn.LogOptions]): Log options.
            timeout (Optional[cdk.Duration]): Execution timeout.

        Returns:
            The created state machine.
        """
        if role is None:
            role = create_role(
                self,
                state_machine_name,
                self.env_base,
                managed_policies=self.required_managed_policies,
                inline_policies_from_statements={
                    "default": self.required_inline_policy_statements,
                },
            )
        else:
            for policy in self.required_managed_policies:
                if isinstance(policy, str):
                    policy = iam.ManagedPolicy.from_aws_managed_policy_name(policy)
                role.add_managed_policy(policy)

            for statement in self.required_inline_policy_statements:
                role.add_to_policy(statement)

        return sfn.StateMachine(
            self,
            self.get_construct_id(state_machine_name),
            state_machine_name=self.env_base.get_state_machine_name(state_machine_name),
            logs=logs or create_log_options(self, state_machine_name, self.env_base),
            role=(
                role if role is not None else create_role(self, state_machine_name, self.env_base)
            ),  # type: ignore[arg-type]
            definition_body=sfn.DefinitionBody.from_chainable(self.definition),
            timeout=timeout,
        )

    @property
    def required_managed_policies(self) -> Sequence[iam.IManagedPolicy | str]:
        """Get required managed policies for this fragment.

        Returns:
            Sequence of required managed policies.
        """
        return []

    @property
    def required_inline_policy_statements(self) -> Sequence[iam.PolicyStatement]:
        """Get required inline policy statements for this fragment.

        Returns:
            Sequence of required policy statements.
        """
        return []


class LazyLoadStateMachineFragment(EnvBaseStateMachineFragment):
    """State machine fragment with lazy-loaded definition.

    The definition is built on first access via build_definition().
    """

    @property
    def definition(self) -> sfn.IChainable:
        """Get the state machine definition, building if needed.

        Returns:
            The chainable definition.
        """
        try:
            return self._definition
        except AttributeError:
            self.definition = self.build_definition()
            return self.definition

    @definition.setter
    def definition(self, value: sfn.IChainable):
        """Set the state machine definition.

        Args:
            value (sfn.IChainable): The definition to set.
        """
        self._definition = value

    @abstractmethod
    def build_definition(self) -> sfn.IChainable:
        """Build the state machine definition.

        Subclasses must implement this method.

        Returns:
            The built definition.
        """
        raise NotImplementedError("Must implement")


class TaskWithPrePostStatus(LazyLoadStateMachineFragment):
    """State machine fragment with pre/post status tracking.

    Wraps a task with status updates for started, failed, and completed states.

    """

    def __init__(
        self,
        scope: constructs.Construct,
        id: str,
        env_base: EnvBase,
        task: sfn.IChainable | None,
    ) -> None:
        """Initialize a task with pre/post status tracking.

        Args:
            scope (constructs.Construct): The construct scope.
            id (str): The construct ID.
            env_base (EnvBase): Environment base for resource naming.
            task (Optional[sfn.IChainable]): The task to wrap.
        """
        super().__init__(scope, id, env_base)
        self.task = task
        self.task_name = id

        self.raw_task_input_path = JsonReferencePath("$")

    @property
    def task(self) -> sfn.IChainable:
        """Get the wrapped task.

        Returns:
            The task.

        Raises:
            AssertionError: If task is not set.
        """
        assert self._task, "Task must be set"
        return self._task

    @task.setter
    def task(self, value: sfn.IChainable | None):
        """Set the wrapped task.

        Args:
            value (Optional[sfn.IChainable]): The task to set.
        """
        self._task = value

    @property
    def task_name(self) -> str:
        """Get the task name.

        Returns:
            The task name.
        """
        return self._task_name

    @task_name.setter
    def task_name(self, value: str):
        """Set the task name.

        Args:
            value (str): The name to set.
        """
        self._task_name = value

    def build_definition(self) -> sfn.IChainable:
        """Build the task definition with status tracking.

        Returns:
            The complete task definition with pre/post status states.
        """
        # Should only evaluate once, otherwise errors for duplicate construct will be raised\
        task__augment_input = self.task__augment_input
        task__status_started = self.task__status_started
        task__status_failed = self.task__status_failed
        task__status_completed = self.task__status_completed
        task__pre_run = self.task__pre_run
        task__post_run = self.task__post_run

        # ---------------------------
        # START DEFINITION
        # ---------------------------
        definition: sfn.Chain | sfn.Pass = sfn.Pass(
            self,
            f"{self.task_name} Start",
            parameters={
                "input": self.raw_task_input_path.as_jsonpath_object,
                "context": self.task_context,
            },
        )

        # -------------
        # AUGMENT INPUT
        # -------------
        if task__augment_input:
            definition = definition.next(
                sfn.Chain.start(task__augment_input).to_single_state(
                    f"{self.task_name} Augment Input",
                    result_path=self.task_input_path.as_reference,
                    output_path=f"{self.task_input_path.as_reference}[0]",
                )
            )

        # -------------
        # PRE TASK
        # -------------
        if task__status_started:
            definition = definition.next(
                sfn.Chain.start(task__status_started).to_single_state(
                    f"{self.task_name} Status Started", result_path=sfn.JsonPath.DISCARD
                )
            )
        if task__pre_run:
            definition = definition.next(
                sfn.Chain.start(task__pre_run).to_single_state(
                    f"{self.task_name} Pre Run", result_path=sfn.JsonPath.DISCARD
                )
            )

        # -------------
        # TASK
        # -------------
        task_chain = sfn.Chain.start(self.task)
        task_enclosure = task_chain.to_single_state(
            f"{self.task_name} Run",
            input_path=self.task_input_path.as_reference,
            result_path=self.task_result_path.as_reference,
        )
        # fmt: off
        definition = (
            definition
            .next(task_enclosure)
            .next(
                sfn.Pass(
                    self,
                    f"{self.task_name} Run (Restructure)",
                    input_path=f"{self.task_result_path.as_reference}[0]",
                    result_path=self.task_result_path.as_reference,
                )
            )
        )
        # fmt: on

        # -------------
        # TASK FAILED
        # -------------
        if task__status_failed:
            task_enclosure.add_catch(
                sfn.Chain.start(task__status_failed)
                .to_single_state(
                    f"{self.task_name} Status Failed", result_path=sfn.JsonPath.DISCARD
                )
                .next(
                    sfn.Fail(
                        self,
                        f"{self.task_name} Fail State",
                        cause=f"Task {self.task_name} failed during execution.",
                    )
                ),
                result_path=self.task_error_path.as_reference,
            )

        # -------------
        # POST TASK
        # -------------
        if task__post_run:
            definition = definition.next(
                sfn.Chain.start(task__post_run).to_single_state(
                    f"{self.task_name} Post Run", result_path=sfn.JsonPath.DISCARD
                )
            )
        if task__status_completed:
            definition = definition.next(
                sfn.Chain.start(task__status_completed).to_single_state(
                    f"{self.task_name} Status Complete", result_path=sfn.JsonPath.DISCARD
                )
            )

        definition.next(
            sfn.Pass(self, f"{self.task_name} End", input_path=self.task_result_path.as_reference)
        )

        # ---------------------------
        # COMPLETE DEFINITION
        # ---------------------------
        return definition

    @property
    def task_input_path(self) -> JsonReferencePath:
        return JsonReferencePath("input")

    @property
    def task_context(self) -> dict[str, Any]:
        return {}

    @property
    def task_result_path(self) -> JsonReferencePath:
        return JsonReferencePath("result")

    @property
    def task_error_path(self) -> JsonReferencePath:
        return JsonReferencePath("error")

    @property
    def task_context_path(self) -> JsonReferencePath:
        return JsonReferencePath("context")

    @property
    def task__augment_input(self) -> sfn.IChainable | None:
        """Run right after the sfn.Pass 'START' of the state machine fragment. Can be used
        to dynamically fill out 'contexts' JsonReferencePath for subsequent lambdas/tasks to use.
        NOTE: Outputs of this chain are MAINTAINED
        """
        return None

    @property
    def task__status_started(self) -> sfn.IChainable | None:
        """Runs right before "task__pre_run" if defined. Otherwise runs right before main
        task executes.
        NOTE: Outputs within this chain get DISCARDED
        """
        return None

    @property
    def task__status_completed(self) -> sfn.IChainable | None:
        """Runs right after "task__post_run" if defined. Otherwise runs right after main
        task is completed.
        NOTE: Outputs within this chain get DISCARDED
        """
        return None

    @property
    def task__status_failed(self) -> sfn.IChainable | None:
        """Runs if main task fails during execution"""
        return None

    @property
    def task__pre_run(self) -> sfn.IChainable | None:
        """Runs right before main task executes.
        NOTE: Outputs within this chain get DISCARDED
        """
        return None

    @property
    def task__post_run(self) -> sfn.IChainable | None:
        """Runs right after main task is completed
        NOTE: Outputs within this chain get DISCARDED
        """
        return None
